
/*  ************************************************************************  *
 *				     main.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <signal.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    <keepc.h>
#include    <standard.h>

static char description [] = "\n\
Allocates extended memory through the int 15h interface.\n\
\n\
  INT15MEM [kilobytes]\n\
\n\
Without an argument, the program reports the current availability.\n";

/*  External references to the resident int 15h handler and two resident
    data variables - all declared in HANDLER.C	*/

extern void _interrupt _far int15h (void);

extern void (_interrupt _far * _RESIDENT lpOrgInt15h) (void);
extern WORD _RESIDENT MemLeftInKB;

/*  The determination of memory available through the int 15h interface is
    set aside for a separate function, since it uses in-line assembler.  */

static WORD GetExtMemInKB (void);

/*  For safety while terminating and staying resident  */

static int _far HardErrorTrap (void)
{
    return (_HARDERR_FAIL);
}

/*  Message output  */

static void PutError (char *str, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register char *ptr;
    WORD AvailMemInKB;
    register WORD MemToAllocateInKB = 0;

    /*	Parse the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    if (ptr [1] == '?' AND ptr [2] == '\0') {
		printf (description);
		return (0);
	    }
	    PutError ("invalid switch %s", ptr);
	    return (-1);
	}
	else {
	    if (MemToAllocateInKB != 0) {
		PutError ("too many parameters %s", ptr);
		return (-1);
	    }
	    MemToAllocateInKB = atoi (ptr);
	    if (MemToAllocateInKB == 0) {
		PutError ("non-zero number expected, not %s", ptr);
		return (-1);
	    }
	}
    }

    /*	Determine the amount of extended memory currently available and
	report it.  If no allocation is also desired, then exit.  Otherwise,
	check that the allocation does not exceed the amount available.  */

    AvailMemInKB = GetExtMemInKB ();
    printf ("\nAmount of extended memory available through int 15h is %dKB\n",
	AvailMemInKB);

    if (MemToAllocateInKB == 0) return (0);

    if (MemToAllocateInKB > AvailMemInKB) {
	PutError ("insufficient extended memory available");
	return (-1);
    }

    /*	Everything's in order.  Now work out how much memory will be left
	behind and hook int 15h so that all future requests for extended
	memory will think that the reduced amount is all that exists.  Do
	not forget to prevent DOS exceptions from terminating the program
	while the int 15h vector is hooked.  */

    signal (SIGINT, SIG_IGN);
    _harderr (HardErrorTrap);

    printf ("Now allocating %dKB of extended memory, ", MemToAllocateInKB);
    MemLeftInKB = AvailMemInKB - MemToAllocateInKB;
    printf ("leaving %dKB behind.\n", MemLeftInKB);

    lpOrgInt15h = _dos_getvect (0x15);
    _dos_setvect (0x15, int15h);

    /*	Tidy up.  */

    CloseAllFiles ();
    FreeEnvironment ();

    /*	Use the ExitKeepC function, to terminate, but note that in this
	simple case, _dos_keep (0, (_segment) lpBreakAddress - _psp) is
	equivalent.  */

    ExitKeepC (0, 0);
}

static WORD GetExtMemInKB (void)
{
    _asm {
	    mov     ah,88h
	    clc
	    int     15h
	    jnc     done
	    xor     ax,ax
	done:
    }

}

static void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nINT15MEM error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

