
;   ************************************************************************
;   *				   driver.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		driver.inc
  INCLUDE		keepsegs.inc
.LIST

.MODEL			SMALL, C

;   The device driver's attribute bits and name may be supplied as symbols
;   at assembly time.  Otherwise, they take default values which could be
;   modified by a utility program after assembly and linkage - this would
;   permit distribution as an object module.  The default values are
;   suitable for programs cast as character device drivers merely to get
;   code into memory as early as possible during system startup.

IFNDEF			_ATTRIBUTE_
  _ATTRIBUTE_		EQU	IS_CHAR_DEVICE
ENDIF

IFNDEF			_NAME_
  _NAME_		TEXTEQU <"">
ELSE
  _NAME_		CATSTR	<!">, _NAME_, <!">
ENDIF

;   ========================================================================

.CODE			KEEP_TEXT

;   Device driver header - this must be located at the very beginning of
;   the device driver:

			PUBLIC	DriverHeader
DriverHeader		DRIVER_HEADER	{ \
					0FFFFFFFFh,
					_ATTRIBUTE_,
					OFFSET KEEP_TEXT:strategy,
					OFFSET KEEP_TEXT:interrupt,
					{_NAME_} \
					}

;   ========================================================================

.FARDATA		KEEP

;   Jump table for device driver commands:

			PUBLIC	DrvrProcTable
DrvrProcTable		LABEL	WORD

;   All procedures in the table have C language prototypes of the form

;	unsigned _near _cdecl DrvrProc (struct REQUEST_HEADER _far *)

;   Functions not provided in other modules default to one which returns the
;   unknown command error, except that initialisation is compulsory.

NEXTNUM 		=	00h
@DrvrProc		MACRO	num:REQ, command:REQ
  IF			num LT NEXTNUM
    .ERR		<@DrvrProc macro not in sequence>
  ENDIF
  IF			num GT NEXTNUM
    REPEAT		num - NEXTNUM
			dw	OFFSET KEEP_TEXT:DrvrProcBadCommand
    ENDM
  ENDIF
  .CODE 		KEEP_TEXT
  IF			num EQ 00h
			EXTERN	DrvrProc&command:NEAR
  ELSE
			EXTERN	DrvrProc&command (DrvrProcBadCommand):NEAR
  ENDIF
  .FARDATA		KEEP
			dw	OFFSET KEEP_TEXT:DrvrProc&command
  NEXTNUM		=	num + 01h
ENDM

@DrvrProc		00h, Init
@DrvrProc		01h, MediaCheck
@DrvrProc		02h, BuildBPB
@DrvrProc		03h, IOCTLRead
@DrvrProc		04h, Read
@DrvrProc		05h, NondestructiveRead
@DrvrProc		06h, InputStatus
@DrvrProc		07h, InputFlush
@DrvrProc		08h, Write
@DrvrProc		09h, WriteWithVerify
@DrvrProc		0Ah, OutputStatus
@DrvrProc		0Bh, OutputFlush
@DrvrProc		0Ch, IOCTLWrite
@DrvrProc		0Dh, OpenDevice
@DrvrProc		0Eh, CloseDevice
@DrvrProc		0Fh, RemovableMedia
@DrvrProc		10h, OutputUntilBusy
@DrvrProc		13h, GenericIOCTL
@DrvrProc		17h, GetLogicalDevice
@DrvrProc		18h, SetLogicalDevice
@DrvrProc		19h, IOCTLQuery

;   ------------------------------------------------------------------------

;   Storage variable for the request header's address:

			PUBLIC	lpRequestHeader
lpRequestHeader 	dd	00000000h

;   ========================================================================

.CODE			KEEP_TEXT

strategy		PROC	FAR PRIVATE USES ds ax

;   Save the address of the request header.

			mov	ax,KEEP_GROUP
			mov	ds,ax
			ASSUME	ds:KEEP_GROUP

			mov	word ptr [lpRequestHeader],bx
			mov	word ptr [lpRequestHeader + 02h],es

;   Provide the means to do more than save the request header's address.
;   This will be needed only very rarely, so leave the responsibility of
;   preserving registers to the DrvrProcStrategy procedure, a suitable
;   prototype for which is

;   void _near _saveregs _cdecl DrvrProcStrategy (struct REQUEST_HEADER _far *)

;   Direct the linker to resolve the external reference with a simple near
;   return if no other module provides the DrvrProcStrategy function.

			EXTERN	DrvrProcStrategy (DrvrProcRetn):NEAR

			push	es
			push	bx
			call	DrvrProcStrategy
			add	sp,04h

			ret

strategy		ENDP

;   ------------------------------------------------------------------------

interrupt		PROC	FAR PRIVATE USES es ds bp di si dx cx bx ax

			mov	ax,KEEP_GROUP
			mov	ds,ax
			ASSUME	ds:KEEP_GROUP

;   Determine which procedure from the jump table (if any) will handle the
;   command.

			les	bx,[lpRequestHeader]
			mov	al,es:(REQUEST_HEADER PTR [bx]).cCommand

			cmp	al,NEXTNUM
			jnb	too_high

			xor	ah,ah
			mov	si,ax
			shl	si,1
			mov	si,[si + DrvrProcTable]

			jmp	@f

too_high:
			mov	si,OFFSET DrvrProcBadCommand

@@:

;   Call the relevant C function, passing the address of the request header.
;   Note that the procedure will be called with ds already addressing the
;   resident data, although in practice functions written in C will probably
;   load ds explicitly.

			push	es
			push	bx
			call	si
			add	sp,04h

;   Each procedure is required to return a value for the device driver
;   status word.

			les	bx,[lpRequestHeader]
			mov	es:(REQUEST_HEADER PTR [bx]).wStatus,ax

			ret

interrupt		ENDP

;   ------------------------------------------------------------------------

DrvrProcBadCommand	PROC	NEAR PUBLIC

			mov	ax,D_ERROR OR D_DONE OR D_BAD_COMMAND

DrvrProcRetn		PROC	NEAR PUBLIC

			ret

DrvrProcRetn		ENDP
DrvrProcBadCommand	ENDP

;   ************************************************************************

END

