
;   ************************************************************************
;   *				  keepinit.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		ascii.inc
  INCLUDE		keepinit.inc
.LIST

;   ========================================================================

.CODE			KEEPCINIT

InitKeepC		PROC	FAR C PUBLIC

			EXTERN	SwapSegments:NEAR
			EXTERN	SwapReferences:NEAR
			EXTERN	C _astart:FAR

;   This is the entry point for an executable using the Keep C scheme for
;   resident programming.  Segments are rearranged in memory before passing
;   control to the usual C run-time entry point defined in the CRT0 module.

;   Preserve those registers that are given some significance by the DOS
;   program loader - most important are the two segment registers ds and es
;   which address the PSP, but note that the two bytes of register ax
;   describe the validity of the FCB fields in the PSP and are actually
;   consulted by some programs.

			push	es
			push	ds
			push	ax

;   Guard against linker options like /PACKCODE by checking that all fences
;   are in fact addressed in the form seg:0000.

			mov	ax,OFFSET NOKEEP_BTEXT
			or	ax,OFFSET KEEP_BDATA
			or	ax,OFFSET NOKEEP_BDATA
			jnz	abort

;   The other procedures involved in the segment swap need access to the
;   original segment addresses and to information about size.  It is
;   convenient to pass both in the KEEPINITINFO structure.  The present
;   implementation restricts the resident data and non-resident code to 64KB
;   each (1000h paragraphs).  Note also that were either area empty, the
;   swap would be superfluous.

			mov	ax,DGROUP
			mov	ds,ax
			mov	si,OFFSET KeepInitInfo
			ASSUME	si:PTR KEEPINITINFO

			mov	dx,NOKEEP_BTEXT
			mov	[si].spOtherCode,dx

			mov	bx,KEEP_BDATA
			mov	[si].spKeepData,bx

			mov	ax,NOKEEP_BDATA
			mov	[si].spOtherData,ax

			sub	ax,bx
			jz	done
			cmp	ax,1000h
			jnb	abort
			mov	[si].KeepDataParas,ax
			mov	cl,04h
			shl	ax,cl
			mov	[si].KeepDataSize,ax

			sub	bx,dx
			jz	done
			cmp	bx,1000h
			jnb	abort
			mov	[si].OtherCodeParas,bx
			shl	bx,cl
			mov	[si].OtherCodeSize,bx

			ASSUME	si:NOTHING

;   Swap the resident data with the non-resident code.

			call	SwapSegments
			jc	abort

;   Search the program's image in memory, adjusting the segment addresses
;   that were filled in from the EXE header when the program was being
;   loaded.

			call	SwapReferences
			jc	abort

done:

;   Pass control to the C run-time as if nothing had ever happened!

			pop	ax
			pop	ds
			pop	es

			jmp	_TEXT:_astart

abort:

;   If for some reason, the segment rearrangement cannot be completed,
;   terminate the program.

			mov	dx,OFFSET KeepInitError
			mov	ah,09h
			int	21h

			mov	ax,4CFFh
			int	21h

InitKeepC		ENDP

;   ========================================================================

.CONST

;   The single fatal error message:

KeepInitError		db	CR, LF,
				"Keep C fatal error:  ",
				"Unable to fixup segment references",
				CR, LF,
				"$"

;   ========================================================================

.DATA?

;   Structure for passing information about the program's segmentation:

			PUBLIC	KeepInitInfo
KeepInitInfo		KEEPINITINFO	<>

;   ************************************************************************

;   Direct the linker to make InitKeepC the entry point to the executable.

END			InitKeepC

