
;   ************************************************************************
;   *				  keepinit.inc				   *
;   ************************************************************************

;   This include file is common to the source files which initialise the
;   Keep C scheme for resident programming in C.

;   The segmentation model for the Keep C scheme is defined in the include
;   file KEEPSEGS.INC and must appear first to ensure correct ordering of
;   segments.

INCLUDE 		keepsegs.inc

;   Initialisation consists of rearranging segments in memory before passing
;   control to the program's usual entry point.  The areas of memory to be
;   swapped are delineated by paragraph-aligned fences:

;	Before				After

;	KEEP_BTEXT			KEEP_BTEXT
;	    resident code		    resident code
;	NOKEEP_BTEXT			KEEP_BDATA
;	    other code			    resident data
;	KEEP_BDATA			NOKEEP_BTEXT
;	    resident data		    other code
;	NOKEEP_BDATA			NOKEEP_BDATA
;	    other data			    other data

;   The resident data will be swapped with the non-resident code.  The
;   procedures responsible for this must be safely outside these regions,
;   which is why the Keep C initialisation code is declared in a special
;   segment in the non-resident data area.

KEEPCINIT		SEGMENT WORD PUBLIC 'KEEPCINIT'
KEEPCINIT		ENDS

;   The default segments:

.MODEL			SMALL, FARSTACK

;   Optional material offering general support for resident programming is
;   supplied in the KEEPCLIB library:  an INCLUDELIB directive here saves
;   the user from specifying that library explicitly when linking the
;   program.

INCLUDELIB		keepclib.lib

;   ========================================================================

;   Structure for passing information about the program's segmentation:

KEEPINITINFO		STRUCT

spOtherCode		dw	?
OtherCodeParas		dw	?
OtherCodeSize           dw      ?

spKeepData		dw	?
KeepDataParas		dw	?
KeepDataSize            dw      ?

spOtherData		dw	?

KEEPINITINFO		ENDS

;   ************************************************************************

