
;   ************************************************************************
;   *				  progname.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		arena.inc
  INCLUDE		keepinit.inc
  INCLUDE		psp.inc
.LIST

;   ========================================================================

.CODE			KEEPCINIT

GetProgramName		PROC	NEAR PUBLIC

;   This procedure expects AX to contain the segment address of a PSP.	It
;   uses this to obtain the pathname with which that process was loaded from
;   disk, returning with a far pointer in dx:ax.

;   After the DOS "load and execute" function 4Bh has constructed an
;   environment for the new process, it appends the pathname whose address
;   was supplied as DS:DX.  The word at offset 2Ch in the PSP is the segment
;   address of the process's environment, but note that it is possible to
;   load a program without an environment, in which case this word will be
;   zero.

			mov	es,ax
			xor	di,di

			mov	dx,es:(PSP PTR [di]).spEnvironment
			or	dx,dx
			jz	badenv

;   The memory arena header immediately below the environment can be used
;   both to verify that the environment block belongs to the designated PSP
;   and to obtain the size of the environment.

			dec	dx
			mov	es,dx
			cmp	es:(ARENA_HEADER PTR [di]).spOwner,ax
			jnz	badenv

			mov	cx,es:(ARENA_HEADER PTR [di]).wParagraphs
			shl	cx,1
                        shl     cx,1
                        shl     cx,1
                        shl     cx,1

			inc	dx
			mov	es,dx

;   The environment consists of contiguous ASCIIZ strings, the last of which
;   is followed by another null byte.

			xor	al,al

@@:
			repnz	scasb
			jnz	badenv

			dec	cx
			jle	badenv

			scasb
			jnz	@b

;   The pathname - another ASCIIZ string - should follow the word 0001h.  If
;   this all fits within the environment block, return with dx:ax addressing
;   the pathname (and with a clear carry flag).

			dec	cx
                        dec     cx
			jle	badenv

			mov	ax,0001h
			scasw
			jnz	badenv

			mov	bx,di
			xor	al,al
			repnz	scasb
			mov	ax,bx
			jz	@f

badenv:

;   If for any reason, the pathname cannot be found, return a null pointer
;   in dx:ax and a set carry flag.

			xor	ax,ax
			mov	dx,ax

			stc

@@:
			ret

GetProgramName		ENDP

;   ************************************************************************

END

