
;   ************************************************************************
;   *				  segdebug.asm				   *
;   ************************************************************************

;   This file is required for the debug version of the CRTKEEPC library.  It
;   checks for the presence of a SEGDEBUG interface and uses this to inform
;   the debugger about the segment rearrangements.

.NOLIST
  INCLUDE		keepinit.inc
.LIST

;   ========================================================================

.CONST

;   Identification string:

Segdebug		db	"SEGDEBUG", 00h

;   ========================================================================

.DATA?

;   The address at which to call the debugger:

SegdebugEntry		dd	?

;   A list of segments whose repositioning has already been conveyed and a
;   pointer into the list:

SegmentList		dw	40h DUP (?)	;  an arbitrary number
NextEntry		dw	?

;   ========================================================================

.CODE			KEEPCINIT

FindSegdebug		PROC	NEAR PUBLIC USES es ds di si

;   If the int 03h trap points to a segment containing the SEGDEBUG
;   identification string at offset 0100h, then the debugger's entry point
;   is at offset 00FBh in that same segment.  Codeview operates similarly,
;   using the identification string CVDEBUG.

			xor	ax,ax
			mov	dx,ax
			mov	es,ax
			mov	es,es:[03h * 4 + 02h]
			mov	di,0100h
			mov	si,OFFSET Segdebug
			mov	cx,SIZEOF Segdebug
			repz	cmpsb
			jnz	@f

			mov	ax,OFFSET SegmentList
			mov	[NextEntry],ax

			mov	ax,00FBh
			mov	dx,es

@@:
			mov	word ptr [SegdebugEntry],ax
			mov	word ptr [SegdebugEntry + 02h],dx

			ret

FindSegdebug		ENDP

;   ------------------------------------------------------------------------

AlertSegdebug		PROC	NEAR PUBLIC USES es di dx cx bx ax

;   If the SEGDEBUG interface is present, then notify it that the segment DX
;   will now be addressed by AX, unless it has been informed already.

;   All registers are preserved:  note that the SEGDEBUG call preserves
;   registers si, di, bp, ds and es.

			cmp	word ptr [SegdebugEntry],0000h
			jz	done

			mov	cx,[NextEntry]
			cmp	cx,OFFSET SegmentList + SIZEOF SegmentList
			jnb	done

			xchg	ax,dx

			push	ds
			pop	es
			mov	di,OFFSET SegmentList
			sub	cx,di
			jz	@f

			shr	cx,1
			repnz	scasw
			jz	done

@@:
			stosw
			mov	[NextEntry],di

;   The SEGDEBUG call takes arguments on the stack.  In the language of C
;   prototypes, this is Segdebug (1, oldseg, newseg).

			push	dx
			push	ax
			mov	ax,0001h
			push	ax
			call	[SegdebugEntry]
			add	sp,06h

done:
			ret

AlertSegdebug		ENDP

;   ************************************************************************

END

