
;   ************************************************************************
;   *				  swaprefs.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		exehdr.inc
  INCLUDE		keepinit.inc
.LIST

;   ========================================================================

.DATA?

;   A 0200h byte buffer to hold a page from the program's EXE header -
;   preceded by a few bytes to anticipate relocation entries that are split
;   between two pages:

			db	04h DUP (?)
ExePageBuffer		db	0200h DUP (?)

;   Structure for obtaining information about the program's segmentation:

			EXTERN	KeepInitInfo:KEEPINITINFO

;   ========================================================================

.CODE			KEEPCINIT

SwapReferences		PROC	NEAR PUBLIC USES es

			EXTERN	GetProgramName:NEAR

;   Check for the presence of a debugger supporting the SEGDEBUG interface -
;   if present, it may be alerted to the segment rearrangements.

IFDEF			SEGDEBUG
			EXTERN	FindSegdebug:NEAR
			call	FindSegdebug
ENDIF

;   Obtain a pathname for the program file.  The procedure which does this
;   expects the PSP segment address in register AX, returning with dx:ax
;   pointing to the pathname.

			mov	ax,es
			call	GetProgramName
			jc	fail

;   Open the file.  Keep the file handle in register bx for the remainder
;   of the procedure.

			push	ds
			xchg	ax,dx
			mov	ds,ax
			mov	ax,3D00h
			int	21h
			pop	ds
			jc	fail

			mov	bx,ax

;   Attempt to read the first 512 bytes of the file, establish that the file
;   is large enough to contain an EXE header and that it does in fact begin
;   with an EXE signature.

			mov	ah,3Fh
			mov	cx,0200h
			mov	dx,OFFSET ExePageBuffer
			int	21h
			jc	close

			cmp	ax,TYPE EXEHDR
			jb	close

			mov	si,dx

			mov	ax,(EXEHDR PTR [si]).wSignature
			cmp	ax,5A4Dh
			jz	@f
			cmp	ax,4D5Ah
			jnz	close

@@:

;   Prepare to examine the EXE header's relocation table.  There must be at
;   least some entries in the relocation table (this source file generates
;   one), which provides at least a cursory check that the program has not
;   been linked with the /EXEPACK option.

			mov	cx,(EXEHDR PTR [si]).wRelocations
			jcxz	close

			add	si,(EXEHDR PTR [si]).pRelocationTable

next:

;   Read the relocation entries one by one until there are none left.  Each
;   entry is a far pointer offset from the program's address frame.  The
;   detail of fetching an entry are handled in a separate procedure, which
;   decrements cx and advances si, reading a new page from the EXE header
;   if necessary.  The entry itself is returned in ax:di.

			jcxz	done

			call	GetEntry
			jc	close

;   Convert the entry to a far pointer and correct the address if it refers
;   to non-resident code or resident data, since these will have been moved.

			add	ax,KEEP_BTEXT
			call	AdjustSegment

;   If the target of the segment reference is one of those which has been
;   moved, then update it.

			mov	es,ax
			mov	ax,es:[di]
			call	AdjustSegment
			jnc	next
			stosw
			jmp	next

close:
			mov	ah,3Eh
			int	21h

fail:
			stc
			jmp	@f

done:
			mov	ah,3Eh
			int	21h

@@:
			ret

SwapReferences		ENDP

;   ------------------------------------------------------------------------

GetEntry		PROC	NEAR PRIVATE

			push	cx

			mov	dx,OFFSET ExePageBuffer

;   In most cases, SI will address the relocation entry which this procedure
;   is intended to return.  The entry will be taken from a buffer containing
;   a page (0200h bytes) from the EXE header.  If there are at least 4 bytes
;   left in the buffer, then read the entry into ax:di and return, having
;   advanced si and decremented cx.  If no space is left in the buffer, then
;   load a new page, reset si to the start of the buffer and continue as
;   before.

			mov	cx,OFFSET ExePageBuffer + 0200h
			sub	cx,si
			cmp	cx,04h
			jnb	read_entry
			or	cx,cx
			jnz	split_entry
			mov	si,dx
			jmp	load_page

split_entry:

;   If the relocation entry is split between two pages, copy the remaining
;   bytes to the area left immediately in front of the buffer (for precisely
;   this purpose), before going on to load a new page.

			push	ds
			pop	es
			mov	di,dx
			sub	di,cx
			mov	ax,di
			rep	movsb
			mov	si,ax

load_page:
			mov	ah,3Fh
			mov	cx,0200h
			int	21h
			jc	@f

read_entry:
			lodsw
			mov	di,ax
			lodsw
			clc

@@:
			pop	cx
			dec	cx
			ret

GetEntry		ENDP

;   ------------------------------------------------------------------------

AdjustSegment		PROC	NEAR PRIVATE

;   Return a clear carry flag unless AX is in the region containing segments
;   that have been swapped, i.e., unless

;	original NOKEEP_BTEXT <= AX < original NOKEEP_BDATA

			cmp	[KeepInitInfo.spOtherCode],ax
			ja	done

			cmp	ax,[KeepInitInfo.spOtherData]
			jnb	done

;   Before correcting the segment value, save the original in dx for the
;   debugger.

IFDEF			SEGDEBUG
			mov	dx,ax
ENDIF

;   Non-resident code has been moved upwards to make way for the resident
;   data.
			cmp	ax,[KeepInitInfo.spKeepData]
			jnb	@f

			add	ax,[KeepInitInfo.KeepDataParas]
			jmp	set

@@:

;   Resident data has been moved downwards.

			sub	ax,[KeepInitInfo.OtherCodeParas]

set:

;   Tell the debugger about the change in segment value so that it may
;   update its symbol tables (pass ax = new value, dx = old value).

IFDEF			SEGDEBUG
			EXTERN	AlertSegdebug:NEAR
			call	AlertSegdebug
ENDIF

			stc

done:
			ret

AdjustSegment		ENDP

;   ************************************************************************

END

