
;   ************************************************************************
;   *				  swapsegs.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		keepinit.inc
.LIST

;   ========================================================================

.DATA?

;   A paragraph buffer used for the segment swap and a counter to determine
;   when the swap has been completed:

buffer			db	10h DUP (?)
count			dw	?

;   Structure for obtaining information about the program's segmentation:

			EXTERN	KeepInitInfo:KEEPINITINFO

;   ========================================================================

.CODE			KEEPCINIT

SwapSegments		PROC	NEAR PUBLIC USES es

;   Resident data is swapped with non-resident code to gather resident
;   material at the base of the program.  The algorithm employed in this
;   procedure avoids any possibly excessive memory allocation by moving data
;   a paragraph at a time, cycling through the transfer region (non-resident
;   code and resident data) until a complete swap has been effected.

;   Each paragraph is moved once, disregarding an occasional need to store a
;   paragraph temporarily.  Counting the number of moves provides an easy
;   way to establish completion.

			mov	dx,[KeepInitInfo.spOtherCode]
			mov	ax,[KeepInitInfo.spOtherData]
			sub	ax,dx
			mov	[count],ax

;   Begin the cycle at offset zero in the non-resident code region.  If 
;   cocycles are necessary, each will be offset one paragraph from its
;   predecessor.
			xor	ax,ax

next:

;   Save the starting address of the current cocycle.

			mov	bx,ax

;   The first paragraph of code is copied to a temporary buffer, then
;   replaced by the first paragraph of data.  A cycle is then established,
;   in which paragraphs are moved into place as space becomes available -
;   the source of the previous transfer becomes the destination of the next.
;   The cycle closes when the address calculated as the next source is the
;   original location of the first paragraph of code.

			push	ds
			pop	es
			mov	di,OFFSET buffer

copy:

;   At this stage, es:di will already have been pointed to the destination,
;   this being (in general) the source of the previous transfer.  The new
;   source will have been calculated in dx:ax - load this into ds:si and
;   copy the data.

			push	ds
			mov	si,ax
			mov	ds,dx
			mov	cx,08h
			rep	movsw
			pop	ds

			dec	[count]

;   Set the destination for the next transfer.

			mov	di,ax
			mov	es,dx

;   Compute the next source.  This means advancing dx:ax by the size of the
;   non-resident code, but is complicated by segmentation and by the need to
;   check for completion of the cycle.	Note that the present implementation
;   assumes that the non-resident code and resident data may each be
;   addressed with respect to a single segment, although together, they may
;   exceed 64KB.

			cmp	dx,[KeepInitInfo.spKeepData]
			jnb	@f

;   If the previous source was in the code segment, the next source will
;   have the same offset but with respect to the data segment instead (since
;   the displacement = codesize), but wrap-around (back to the code segment
;   and requiring an adjustment of datasize) occurs if ax >= datasize.

			cmp	ax,[KeepInitInfo.KeepDataSize]
			jnb	wraparound

			mov	dx,[KeepInitInfo.spKeepData]

			jmp	copy

@@:

;   The source pointer (currently in the data region) is advanced by the
;   size of non-resident code, but note that wrap-around occurs if
;   ax + codesize >= datasize.

			add	ax,[KeepInitInfo.OtherCodeSize]
			jbe	@f

			cmp	ax,[KeepInitInfo.KeepDataSize]
			jb	copy

@@:
			mov	dx,[KeepInitInfo.spOtherCode]

wraparound:

			sub	ax,[KeepInitInfo.KeepDataSize]

;   Because the source for the next transfer lies in the code segment,
;   check for cocycle completion before proceeding.  If this is the last
;   paragraph in the cocycle, fill the request from the temporary storage
;   buffer on the stack.

			cmp	ax,bx
			jnz	copy

			mov	si,OFFSET buffer

			mov	cx,08h
			rep	movsw

;   If the number of paragraphs of non-resident code and resident data are
;   not relatively prime, then a complete swap requires more than one
;   cocycle of paragraph transfers, each offset by one paragraph from the
;   one before.

			add	ax,10h

			cmp	[count],0000h
			jnz	next

			ret

SwapSegments		ENDP

;   ************************************************************************

END

