
;   ************************************************************************
;   *				  keepexit.asm				   *
;   ************************************************************************

.NOLIST
  INCLUDE		keepsegs.inc
  INCLUDE		psp.inc
.LIST

.MODEL			SMALL, FARSTACK

;   ========================================================================

.DATA?

;   An uninitialised array in which to store addresses of functions
;   registered by AtExitKeepC or OnExitKeepC:

			PUBLIC	XKEEPC_start, XKEEPC_end

XKEEPC_start		dd	10h DUP (?)
XKEEPC_end		LABEL	DWORD

;   ========================================================================

.DATA

;   A pointer into the array XKEEPC_start:

			PUBLIC	XKEEPC_ptr
XKEEPC_ptr		dw	OFFSET XKEEPC_start

;   Make the default break address available publicly.

			PUBLIC	C lpBreakAddress
lpBreakAddress		dd	NOKEEP_BTEXT:0000h

;   The amount of memory for retention is calculated with respect to the
;   PSP's segment address - supplied by the C run-time:

			EXTERN	C _psp:WORD

;   ========================================================================

.CODE

ExitKeepC		PROC	C PUBLIC retcode:WORD, breakaddr:DWORD

IF			@CodeSize
			EXTERN	C _cexit:FAR
ELSE
			EXTERN	C _cexit:NEAR
ENDIF

;   int _cdecl ExitKeepC (unsigned, void _far *)

;   Before examining the caller's requested break address, check the amount
;   of memory available in the program's main block.  This is not done just
;   for defensiveness - it is the programmer's responsibility to ensure that
;   his requested break address is sensible - but because resident
;   pre-terminators are called from code relocated to the top of memory
;   available to the program.

;   DOS function 4Ah, if passed an impossible value in register bx, will
;   return with bx corrected to the largest valid size in paragraphs.  That
;   it also expands the block even though the function fails is the subject
;   of one of those interminable debates about the difference between bugs
;   and features.

			mov	ah,4Ah
			mov	bx,0FFFFh
			mov	es,[_psp]
			int	21h

;   Update the memory ceiling field in the PSP.

			mov	ax,es
			add	ax,bx
			mov	es:[PSP.spCeiling],ax

;   Retention just of the Keep C code and data segments may be specified by
;   supplying a null pointer for the break address.  If the break address is
;   not aligned to a paragraph boundary, advance it to the next one.  In all
;   cases, arrange to have the break segment (the segment address of the
;   lowest unneeded paragraph) in register dx.

			mov	ax,word ptr [breakaddr]
			mov	dx,word ptr [breakaddr + 02h]
			or	ax,ax
			jnz	normalise
			or	dx,dx
			jnz	@f
			mov	dx,NOKEEP_BTEXT
			jmp	@f
normalise:
			add	ax,0Fh
			rcr	ax,1
			shr	ax,1
                        shr     ax,1
                        shr     ax,1
			add	dx,ax
@@:

;   Perform the normal (non-resident) clean up.  This includes any user
;   terminators registered with the usual atexit or onexit functions.
;   Preserve the break segment across the call.

			push	dx
			call	_cexit
			pop	dx

;   To support programs that need to overwrite their initialisation code
;   with data before terminating (as opposed to programs which want merely
;   to reserve the memory), the AtExitKeepC and OnExitKeepC functions are
;   provided to register as many as 16 (an arbitrary number) other functions
;   in a resident code segment to be called from safe memory before the
;   program terminates.  Matters are of course much easier if no such
;   functions have been registered.

			mov	di,[XKEEPC_ptr]
			cmp	di,OFFSET XKEEPC_start
			jnz	@f

;   Subtract the PSP's segment address and use DOS function 31h to
;   terminate.

			sub	dx,[_psp]
			mov	al,byte ptr [retcode]

			jmp	terminate

@@:

;   Although the pre-terminators must lie in the resident code segment to
;   guarantee their own safety, it is possible that the address table and
;   even the stack may be overwritten if for example the program wants to
;   clear a very large data array before terminating.

;   Whether the table and stack need to be relocated or not, the procedure
;   for calling the pre-terminators expects es:si to address the table's
;   base and es:di to provide an upper bound (which it does already!).

			push	ds
			pop	es
			mov	si,OFFSET XKEEPC_start

;   Since the stack may be reset, retrieve the program's exit code.

			mov	bx,[retcode]

;   Remember that the lowest unneeded paragraph is in dx.  If this is below
;   or equal to the paragraph in which the address table begins, then the
;   table and stack will be safe.

			mov	ax,OFFSET XKEEPC_start
			mov	cl,04h
			shr	ax,cl
			mov	cx,ds
			add	ax,cx

			cmp	dx,ax
			jbe	ok

;   If the caller wants to overwrite so much memory before going resident
;   that there isn't at least a little room (400h bytes) for the final
;   manipulations, something is clearly wrong, for at the very least, there
;   would be no need to use the Keep C scheme.	In such a case, return under
;   the hopeful assumption that the caller will know what to do.  This is
;   the only cause for ExitKeepC to return.

			mov	es,[_psp]
			mov	ax,es:[PSP.spCeiling]
			sub	ax,dx
			jb	noexit
			cmp	ax,400h SHR 4
			jnb	@f

noexit:
			xor	ax,ax
			ret

@@:

;   Set up a new stack before moving the address table.

			cmp	ax,1000h
			jb	@f
			mov	ax,0FFFh
@@:
			shl	ax,cl

			mov	es,dx
			cli
			mov	ss,dx
			mov	sp,ax
			sti

;   Copy the address table.  Note that it is being moved upwards in memory
;   and just conceivably, the source and destination may overlap, so copy
;   from top to bottom.  The new table will begin at offset zero in the new
;   stack segment.  As before, es:si must address the table's base and es:di
;   its upper bound.

			mov	cx,di
			sub	cx,si
			mov	si,di
			dec	si
                        dec     si
			mov	di,cx
			push	di
			dec	di
                        dec     di
			shr	cx,1
			std
			rep	movsw
			cld
			pop	di
			xor	si,si

ok:

;   Push the two arguments required by DOS function 31h onto the stack - the
;   number of paragraphs first, then the program's exit code.

			sub	dx,[_psp]
			push	dx
			push	bx

;   Append the termination code to the address table - wherever the table is,
;   the memory immediately after it should be safe.

			push	di
			push	si
			push	cs
			pop	ds
			mov	si,OFFSET moveable_exit
			mov	cx,OFFSET endexit - OFFSET moveable_exit
			rep	movsb
			pop	si
			pop	di

;   Set ds to the default group for resident data and transfer control to
;   safe memory.

			mov	ax,KEEP_GROUP
			mov	ds,ax
			ASSUME	ds:KEEP_GROUP

			push	es
			push	di
			retf

moveable_exit:

;   Code from here onwards is executed after relocation away from material
;   that may be overwritten by the pre-terminators.

;   At least one DOS function remains to be called - for TSR termination, if
;   nothing else.  This will be susceptible to Ctrl-Break and Ctrl-C unless
;   the break flag is disabled or an int 23h handler has been installed to
;   prevent premature (non-resident) termination.  The trouble is that if
;   the break-checking was on when the program started, but has since been
;   disabled, then the program should turn it back on before termination.
;   This restoration would be the responsibility of the _cexit cleanup
;   function.  Under no circumstances should int 21h function 31h be used
;   to terminate without ensuring that int 23h exceptions will be returned
;   harmlessly (the int 23h vector will be restored by the DOS kernel - to
;   an address retrieved from the PSP).  It is probably best to arrange the
;   installation of a trivial int 23h vector here - just to be absolutely
;   sure.  Note that using int 21h function 25h to install the Ctrl-C trap
;   will risk premature termination unless break-checking is disabled first.

			mov	ax,3300h
			int	21h
			push	dx

			xor	dx,dx
			mov	ax,3301h
			int	21h

			push	ds
			push	cs
			pop	ds
			mov	dx,OFFSET int23h_atexit - OFFSET moveable_exit
			add	dx,di
			mov	ax,2523h
			int	21h
			pop	ds

			pop	dx
			mov	ax,3301h
			int	21h

;   It is conceivable that a resident pre-terminator may attempt file I/O
;   and generate a critical error.  Ensure that this is defeated by
;   installing a suitable int 24h handler.  Again, the DOS kernel restores
;   the vector as part of the program's termination.  If a Ctrl-C is
;   pending when int 21h function 25h is used this time, then it will be
;   returned harmlessly.

			push	ds
			mov	dx,OFFSET int24h_atexit - OFFSET moveable_exit
			add	dx,di
			mov	ax,2524h
			int	21h
			pop	ds

;   Call the functions in the address table.  As with the pre-terminators
;   registered through the usual atexit and onexit functions, these are
;   called in a last-in first-out order.  The pre-terminators are expected
;   to be far functions lying in a resident code segment.

@@:
			sub	di,04h

			push	es
			call	dword ptr es:[di]
			pop	es

			cmp	si,di
			jb	@b

;   Retrieve the paragraph and exitcode values from the stack, then call DOS
;   function 31h to terminate.

			pop	ax
			pop	dx

terminate:
			mov	ah,31h
			int	21h

int24h_atexit:
			mov	al,03h
int23h_atexit:
			iret

endexit:

ExitKeepC		ENDP

;   ************************************************************************

END

