
/*  ************************************************************************  *
 *				   storage.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <signal.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    <standard.h>

#include    "storage.h"

BOOL ArrangeStorage (DWORD);
WORD ReadStorage (void _far *, WORD);
WORD WriteStorage (void _far *, WORD);
BOOL RewindStorage (void);
void DestroyStorage (void);

extern struct STORAGE_MEDIUM MemStorage;
extern struct STORAGE_MEDIUM XMSStorage;

static struct STORAGE_MEDIUM *storage [] = {&MemStorage, &XMSStorage};
#define STORAGE_MEDIA  (sizeof (storage) / sizeof (storage [0]))

static void CtrlCtrap (int);
static int _far harderrortrap (void);

static BOOL traps_set = FALSE;

extern void PutError (char *, ...);

/*  ========================================================================  */

BOOL ArrangeStorage (DWORD requirement)
{
    register struct STORAGE_MEDIUM **smptr;
    register struct STORAGE_MEDIUM *ptr;
    WORD n;
    BOOL result = FALSE;

    if (requirement == 0) return (FALSE);

    if (NOT traps_set) {
	atexit (DestroyStorage);
	signal (SIGINT, CtrlCtrap);
	_harderr ((void (_far *) ()) (void CODESEG *) harderrortrap);
	traps_set = TRUE;
    }

    smptr = storage, n = STORAGE_MEDIA;
    do {
	ptr = *smptr ++;
	ptr -> capacity = ptr -> create (requirement);
	if (ptr -> capacity == 0) continue;

	ptr -> flags |= SELECTED;

	requirement -= ptr -> capacity;

	if (requirement == 0) {
	    result = TRUE;
	    break;
	}
    } while (-- n);

    while (-- smptr >= storage) {
	ptr = *smptr;
	if (NOT (ptr -> flags & SELECTED)) ptr -> destroy ();
    }

    return (result);
}

WORD WriteStorage (void _far *buffer, WORD count)
{
    register struct STORAGE_MEDIUM **smptr;
    register struct STORAGE_MEDIUM *ptr;
    WORD n;
    WORD numwrt, part_count;
    DWORD space;

    numwrt = 0;
    if (count == 0) return (numwrt);

    smptr = storage, n = STORAGE_MEDIA;
    do {
	ptr = *smptr ++;
	if (ptr -> flags & SELECTED AND NOT (ptr -> flags & FILLED)) {

	    space = ptr -> capacity - ptr -> current;
	    part_count = space < count ? (WORD) space : count;

	    if (ptr -> write (buffer, part_count) != part_count)
		return (FALSE);

	    ptr -> current += part_count;
	    if (ptr -> current >= ptr -> capacity) ptr -> flags |= FILLED;

	    (BYTE _far *) buffer += part_count;
	    numwrt += part_count;
	    count -= part_count;

	    if (count == 0) break;
	}
    } while (-- n);

    return (numwrt);
}

BOOL RewindStorage (void)
{
    register struct STORAGE_MEDIUM **smptr;
    register struct STORAGE_MEDIUM *ptr;
    WORD n;

    smptr = storage, n = STORAGE_MEDIA;
    do {
	ptr = *smptr ++;
	if (ptr -> flags & SELECTED) {
	    if (NOT (ptr -> rewind ())) return (FALSE);
	    ptr -> flags &= ~RETRIEVED;
	    ptr -> current = 0;
	}
    } while (-- n);

    return (TRUE);
}

WORD ReadStorage (void _far *buffer, WORD count)
{
    register struct STORAGE_MEDIUM **smptr;
    register struct STORAGE_MEDIUM *ptr;
    WORD n;
    WORD numread, part_count;
    DWORD space;

    numread = 0;
    if (count == 0) return (numread);

    smptr = storage, n = STORAGE_MEDIA;
    do {
	ptr = *smptr ++;
	if (ptr -> flags & SELECTED AND NOT (ptr -> flags & RETRIEVED)) {

	    space = ptr -> capacity - ptr -> current;
	    part_count = space < count ? (WORD) space : count;

	    if (ptr -> read (buffer, part_count) != part_count) return (FALSE);

	    ptr -> current += part_count;
	    if (ptr -> current >= ptr -> capacity) ptr -> flags |= RETRIEVED;

	    (BYTE _far *) buffer += part_count;
	    numread += part_count;
	    count -= part_count;

	    if (count == 0) break;
	}
    } while (-- n);

    return (numread);
}

void DestroyStorage (void)
{
    register struct STORAGE_MEDIUM **smptr;
    register struct STORAGE_MEDIUM *ptr;
    WORD n;

    smptr = storage, n = STORAGE_MEDIA;
    do {
	ptr = *smptr ++;
	ptr -> destroy ();
    } while (-- n);
}

static void CtrlCtrap (int sigint)
{
    signal (SIGINT, SIG_IGN);
    PutError ("terminated by user");
    exit (2);
}

static int _far harderrortrap (void)
{
    return (_HARDERR_FAIL);
}

/*  ************************************************************************  */

