
/*  ************************************************************************  *
 *				   xmsfile.c				      *
 *  ************************************************************************  */

/*  This file initialises and co-ordinates modules from the XMSFILE device
    driver in order to supply the DISKDUP utility with extended memory.  */

#include    <driver.h>
#include    <smartdrv.h>
#include    <standard.h>

#include    "storage.h"

static DWORD CreateXMS (DWORD);
static BOOL DestroyXMS (void);
static WORD ReadXMS (void _far *, WORD);
static WORD WriteXMS (void _far *, WORD);
static BOOL RewindXMS (void);

struct STORAGE_MEDIUM XMSStorage =
    {0, 0, 0, CreateXMS, DestroyXMS, ReadXMS, WriteXMS, RewindXMS};

/*  The usual access to the XMS server	*/

extern void (_far *GetXMSAddress (void)) (void);
void (_far *lpXMS) (void);

/*  XMSFILE is driven through the DOS device driver interface.	This just
    means that information is passed to and fro in a request header.  */

extern WORD DrvrProcRead (struct REQUEST_HEADER _far *);
extern WORD DrvrProcWrite (struct REQUEST_HEADER _far *);
extern WORD DrvrProcIOCTLWrite (struct REQUEST_HEADER _far *);
extern WORD DrvrProcOpenDevice (struct REQUEST_HEADER _far *);
extern WORD DrvrProcCloseDevice (struct REQUEST_HEADER _far *);

static struct REQUEST_HEADER rqhdr;

/*  Two functions that undercut the device driver interface in XMSFILE but
    which are useful here.  */

extern DWORD AdjustRequest (BYTE, DWORD);
extern WORD Destroy (void);

/*  XMSFILE may take extended memory from the SMARTDRV.SYS disk cache.	It
    needs some functions from SMARTDRV.C in order to detect SMARTDrive's
    presence and suitability.  A flag is left in TRANSFER.C so that the
    XMSFILE device driver code knows to steal memory from the disk cache.  */

extern int OpenSMARTDrive (void);
extern BOOL SupportsIOCTL (int);
extern BOOL ReadSMARTDrive (int, struct SD_READ *);
extern BOOL CloseSMARTDrive (int);

extern BOOL bUseSMARTDrive;

static DWORD CreateXMS (DWORD requirement)
{
    DWORD avail;
    WORD status;
    int sdhandle;
    struct SD_READ sd_read;
    WORD largestavailEMBinKB;
    DWORD largestavailEMB;

    /*	Get the XMS entry point and record it for the routines in the
	XMSFILE modules.  */

    lpXMS = GetXMSAddress ();
    if (lpXMS == NULL) return (0);

    /*	Find the SMARTDRV.SYS disk cache and check that it uses extended
	memory.  */

    sdhandle = OpenSMARTDrive ();
    if (sdhandle != -1) {
	if (SupportsIOCTL (sdhandle)) {
	    if (ReadSMARTDrive (sdhandle, &sd_read)) {
		if (sd_read.MemoryType == 1) bUseSMARTDrive = TRUE;
	    }
	}
	CloseSMARTDrive (sdhandle);
    }

    /*	Open the XMSFILE "device" then ask it how much of the requirement
	could it meet (in the way of memory it can write to).  */

    status = DrvrProcOpenDevice (&rqhdr);
    if (status & D_ERROR) return (0);

    avail = AdjustRequest (D_WRITE, requirement);

    /*	The XMSFILE "device" was given no means to indicate its ability to
	meet part of a request:  its allocations are "all or nothing".  If
	it turns out that the XMSFILE algorithm cannot meet the full
	requirement, scale down the request to the size of the largest free
	EMB and try again.  This is, of course, a hack that will usually
	handle the situation - but for more sophistication and greater
	coverage, it would be better to improve the routines in XMSFILE.  */

    if (avail == 0) {
	_asm {
		mov	ah,08h
		call	lpXMS
		mov	largestavailEMBinKB,ax
	}
	largestavailEMB = largestavailEMBinKB << 10;
	avail = AdjustRequest (D_WRITE, largestavailEMB);
    }
    return (avail);
}

static BOOL DestroyXMS (void)
{
    if (lpXMS != NULL) Destroy ();
    return (TRUE);
}

/*  The remaining functions simply phrase the read, write and rewind
    services for DISKDUP storage into terms that XMSFILE understands.  */

static WORD ReadXMS (void _far *buffer, WORD count)
{
    rqhdr.lpBuffer = buffer;
    rqhdr.wCount = count;
    DrvrProcRead (&rqhdr);
    return (rqhdr.wCount);
}

static WORD WriteXMS (void _far *buffer, WORD count)
{
    rqhdr.lpBuffer = buffer;
    rqhdr.wCount = count;
    DrvrProcWrite (&rqhdr);
    return (rqhdr.wCount);
}

static BOOL RewindXMS (void)
{
    DWORD temp = 0;
    WORD status;

    rqhdr.lpBuffer = (BYTE _far *) &temp;
    rqhdr.wCount = sizeof (temp);
    status = DrvrProcIOCTLWrite (&rqhdr);
    return (status & D_ERROR ? FALSE : TRUE);
}

/*  ************************************************************************  */

