
/*  ************************************************************************  *
 *				   drvinfo.c				      *
 *  ************************************************************************  */

#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <cds.h>
#include    <dpb.h>
#include    <driver.h>
#include    <standard.h>
#include    <sysvars.h>

/*  Some drive features are determined externally.  */

extern BYTE GetBIOSDrive (WORD);
extern struct DPB _far *GetDPB (WORD);
extern void DescribeCDSProperties (WORD);

/*  For message output	*/

void PutComment (CHAR *, ...);
static BOOL firstcomment;

/*  ========================================================================  */

/*  C language equivalents for int 21h functions 4408h and 4409h.  Like the
    DOS functions themselves, these expect 1-based drive numbers.  They
    follow the standard scheme of returning a DOS error code as the function
    value and passing other information at an address supplied as the last
    argument.  */

static WORD _dos_getdrivetype (WORD drive, WORD *ptr)
{
    WORD temp;

    _asm {
	    mov     bx,drive
	    mov     ax,4409h
	    int     21h
	    jc	    done
	    mov     temp,dx
    }
    *ptr = temp;
    _asm {
	    xor     ax,ax
	done:
    }
}

static WORD _dos_removablemedia (WORD drive, WORD *ptr)
{
    WORD temp;

    _asm {
	    mov     bx,drive
	    mov     ax,4408h
	    int     21h
	    jc	    done
	    mov     temp,ax
    }
    *ptr = temp;
    _asm {
	    xor     ax,ax
	done:
    }
}

/*  ------------------------------------------------------------------------  */

BOOL IsDriveDefined (WORD drive)
{
    WORD type;

    /*	If int 21h function 4409h returns an error, then the drive is
	unknown to the DOS kernel.  This is the simplest possible way to
	establish a drive's current existence.  */

    return (_dos_getdrivetype (drive + 1, &type) == 0 ? TRUE : FALSE);
}

/*  The next function lists drive properties.  It may assume that the drive
    number given as the first argument is currently in control of its
    physical device, this having been arranged by the caller, if necessary.
    The second argument gives the drive that was originally in control of
    the physical device.  */

void DescribeDrive (WORD drive, WORD activedrive)
{
    WORD biosdrive, driveflags, errorcode, temp;
    struct DPB _far *dpbptr;

    errorcode = _dos_getdrivetype (drive + 1, &driveflags);
    if (errorcode) return;

    /*	The drive exists, so list it.  The first comment (if any) about the
	drive will follow on the same line.  Subsequent comments are written
	on new lines - this is taken care of by the PutComment function,
	which should be used for all remarks about the drive.  */

    printf ("\n%c:  ", drive + 'A');
    firstcomment = TRUE;

    /*	DOS function 4409h returns flags in register dx.  If the bit masked
	by 0x1000 is set, then all access to the drive is handled via the
	network redirector interface.  */

    if (driveflags & 0x1000) {
	PutComment ("remote/network drive");
	return;
    }

    /*	Local drives may divided into those regarded as physical at DOS
	kernel level because they correspond directly to a unit of some
	block device driver and virtual drives that exist only as artifacts
	of the directory system (e.g., a drive created with the SUBST
	utility).  Physical drives have their own DPB structure.  */

    dpbptr = GetDPB (drive);
    if (dpbptr != NULL AND (WORD) dpbptr -> cDrive == drive) {

	/*  If the drive is a phantom (i.e., if it shares a physical device
	    with some other drive and is not currently in control of that
	    device), then just note the correspondence.  */

	if (activedrive != 0xFF AND activedrive != drive)
	    PutComment ("sharing device with drive %c", activedrive + 'A');
	else {

	    /*	Begin by describing the physical storage device to which the
		DOS drive refers.  For the first attempt, try to identify
		the drive number number used for int 13h access.  */

	    biosdrive = GetBIOSDrive (drive);
	    if (biosdrive != 0xFF) {
		if (biosdrive >= 0x80)
		    PutComment ("hard disk %d", biosdrive - 0x80 + 1);
		else PutComment ("floppy disk drive %d", biosdrive + 1);
	    }
	    else {

		/*  It may be that int 13h is not used to control the
		    storage device.  In many cases, such as the increasingly
		    common CD-ROM drives, documented interfaces exist
		    through which this program could obtain information if
		    it were being developed further.

		    Ironically, the most difficult medium to be certain of
		    categorising correctly may well be the ubiquitous RAM
		    disk:  no interface is provided, but a reasonable test
		    (employed by several Microsoft products) seems to be
		    that only with a safe medium like RAM would a drive with
		    an FAT system risk having only one FAT copy.  */

		if (_dos_removablemedia (drive + 1, &temp) != 0) temp = -1;

		if (temp != 0 AND dpbptr -> cFATs == 1)
		    PutComment ("RAM disk");
		else PutComment ("unknown %s drive",
				    temp == 0 ? "removable" : "fixed");
	    }
	}
    }
    DescribeCDSProperties (drive);
}

void PutComment (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    if (NOT firstcomment) printf ("\n    ");
    else firstcomment = FALSE;
    vprintf (str, argptr);
    va_end (argptr);
}

/*  ************************************************************************  */

