
/*  ************************************************************************  *
 *				    drvmap.c				      *
 *  ************************************************************************  */

#include    <standard.h>

/*  DOS-level functions to control logical drive mapping.  These follow the
    usual scheme of returning a DOS error code (or zero for success)
    directly and recording other information at a specified address.  The
    drive numbers accepted by these functions are 1-based.  */

static WORD _dos_getlogicalmap (WORD drive, BYTE *mapptr)
{
    BYTE temp;

    _asm {
	    mov     bx,drive
	    mov     ax,440Eh
	    int     21h
	    jc	    done
	    mov     temp,al
    }
    *mapptr = temp;
    _asm {
	    xor     ax,ax
	done:
    }
}

static WORD _dos_setlogicalmap (WORD drive)
{
    _asm {
	    mov     bx,drive
	    mov     ax,440Fh
	    int     21h
	    jc	    done
	    xor     ax,ax
	done:
    }
}

/*  ------------------------------------------------------------------------  */

/*  Functions one level higher (and which work with 0-based drive numbers).
    These are interested only in non-trivial mapping of single drives.	*/

BYTE GetLogicalDrive (WORD drive)
{
    BYTE map;

    /*	If it can be established definitely that the specified drive shares
	its physical device with other drives, return the 0-based number of
	the drive currently in control of that physical device.  */

    if (_dos_getlogicalmap (drive + 1, &map) == 0 AND map != 0)
	return (map - (BYTE) 1);

    /*	In any other circumstance, return the value 0xFF.  */

    return (0xFF);
}

/*  The next function tells the responsible device driver to regard the
    designated 0-based drive as the one in charge, should it be sharing its
    physical device with other drives.	*/

BOOL SetLogicalDrive (register WORD drive)
{
    BYTE map;

    /*	If the logical map facility is not supported for the drive, assume
	this is because it is not needed.  */

    if (_dos_getlogicalmap (drive + 1, &map)) return (TRUE);

    /*	If the drive is known not to share its physical device, then again,
	the logical drive is already set.  */

    if (map == 0) return (TRUE);

    /*	If the drive is already the active drive for its shared device, then
	nothing more need be done.  */

    if ((WORD) map == drive + 1) return (TRUE);

    /*	Otherwise, put the new drive in charge.  Failure here, given that
	the device driver supports the Get Logical Drive Map facility, is
	best regarded as an error.  */

    return (_dos_setlogicalmap (drive + 1) == 0 ? TRUE : FALSE);
}

/*  ------------------------------------------------------------------------  */

/*  The following two functions work with the complete drive map.  */

void SaveDriveMap (register BYTE *drivemap)
{
    register WORD drive;

    for (drive = 0; drive < 26; drive ++) {
	*drivemap ++ = GetLogicalDrive (drive);
    }
}

void RestoreDriveMap (register BYTE *drivemap)
{
    register WORD drive;

    for (drive = 0; drive < 26; drive ++) {
	if ((WORD) *drivemap ++ == drive) SetLogicalDrive (drive);
    }
}

/*  ************************************************************************  */

