
/*  ************************************************************************  *
 *		     main.c - primary file for drvinfo.exe		      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <signal.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <standard.h>

static CHAR description [] = "\n\
Presents information about DOS drives\n\
\n\
  DRVINFO [[drive:]...]\n\
";

extern BOOL IsDriveDefined (WORD);
extern void DescribeDrive (WORD, WORD);

extern void RestoreDriveMap (BYTE *);
extern void SaveDriveMap (BYTE *);
extern BOOL SetLogicalDrive (WORD);

static BYTE drivemap [26];

static void cleanup (void);
static void CtrlCTrap (int);
static int _far HardErrorTrap (void);

void PutError (CHAR *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register CHAR *ptr;
    register WORD drive;
    BOOL listdrives;
    BYTE infowanted [26];

    if (MK_WORD (_osmajor, _osminor) < MK_WORD (3, 20)) {
	PutError ("requires DOS version 3.20 or higher");
	return (-1);
    }

    /*	Parse the command line.  The chief objective is to read a list of
	drives about which information is sought.  */

    memset (infowanted, FALSE, sizeof (infowanted));
    listdrives = FALSE;

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;
	    if (ptr [0] == '?' AND ptr [1] == '\0') {
		printf (description);
		return (0);
	    }
	    PutError ("invalid switch - %s", -- ptr);
	    return (-1);
	}
	else {
	    drive = (ptr [0] & ~0x20) - 'A';
	    if (drive <= 'Z' - 'A' AND ptr [1] == ':' AND ptr [2] == '\0') {
		infowanted [drive] = TRUE;
		listdrives = TRUE;
		continue;
	    }
	    PutError ("invalid parameter - %s", ptr);
	    return (-1);
	}
    }

    /*	If the command line contained no drive specifications, assume that
	information is required for all drives.  */

    if (NOT listdrives) memset (infowanted, TRUE, sizeof (infowanted));

    /*	Take undefined drives out of the list and exit from the program if
	no valid drives remain.  */

    listdrives = FALSE;
    for (drive = 0; drive < 26; drive ++) {
	if (infowanted [drive]) {
	    infowanted [drive] = (BYTE) IsDriveDefined (drive);
	    if (infowanted [drive]) listdrives = TRUE;
	}
    }
    if (NOT listdrives) {
	PutError ("no valid drives in list");
	return (-1);
    }

    /*	Save the logical drive map and make the necessary provisions to
	ensure restoration when the program terminates (or is terminated).
	A reasonable course would be to inhibit critical error generation
	and also defeat the fatal action of Ctrl-C, but it is only a small
	effort to support orderly termination via Ctrl-C.  */

    #ifndef _NO_DRIVE_MAP_
    SaveDriveMap (drivemap);
    #endif

    atexit (cleanup);
    signal (SIGINT, CtrlCTrap);
    _harderr (HardErrorTrap);

    /*	For each drive for which a description has been requested, call an
	external function to get information and display it:  this division
	means that the drive-by-drive description can be changed in nature
	without affecting this module.	*/

    for (drive = 0; drive < 26; drive ++) {
	if (infowanted [drive]) {

	    /*	It eases matters greatly that the external function be
		entitled to assume that the drive in question actually has
		logical control of its physical device.  If the drive shares
		its physical device with other logical drives and does not
		happen to be the one currently in control, alter the logical
		drive map before commencing any investigation.	*/

	    #ifndef _NO_DRIVE_MAP_
	    SetLogicalDrive (drive);
	    DescribeDrive (drive, drivemap [drive]);
	    #else
	    DescribeDrive (drive, 0xFF);
	    #endif
	}
    }

    printf ("\n");
    return (0);
}

static void cleanup (void)
{
    #ifndef _NO_DRIVE_MAP_
    RestoreDriveMap (drivemap);
    #endif
}

static void CtrlCTrap (int sigint)
{
    signal (SIGINT, SIG_IGN);
    PutError ("terminated by user");
    exit (-1);
}

static int _far HardErrorTrap (void)
{
    return (_HARDERR_FAIL);
}

void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nDRVINFO fatal error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

