
;   ************************************************************************
;   *				  keepsegs.inc				   *
;   ************************************************************************

;   Segmentation model for a C program (TSR or device driver) that discards
;   code and data needed only for initialisation.

;   Paragraph-aligned fence marking the start of the resident code region:

KEEP_BTEXT		SEGMENT PARA COMMON 'CODE'
KEEP_BTEXT		ENDS

;   Resident code segment:

;	The command line switch /NT KEEP_TEXT directs the Microsoft C
;	compiler to place all of a module's code in a segment KEEP_TEXT
;	with class CODE.  The present implementation of the Keep C scheme
;	provides no other segment for resident code.

KEEP_TEXT		SEGMENT WORD PUBLIC 'CODE'
KEEP_TEXT		ENDS

;   Paragraph-aligned fence for non-resident code segments:

NOKEEP_BTEXT		SEGMENT PARA COMMON 'CODE'
NOKEEP_BTEXT		ENDS

;   All other segments belonging to class 'CODE', such as the default code
;   segment _TEXT, will be placed here.  Given DOSSEG ordering, this area
;   will also include those segments whose classnames end in 'CODE'.

;   Paragraph-aligned fence marking the start of the resident data region:

KEEP_BDATA		SEGMENT PARA COMMON 'FAR_DATA'
KEEP_BDATA		ENDS

;   Resident data segments:

;	Supplying the command line switch /ND KEEP to the C compiler results
;	in the creation of the segments KEEP, KEEP_CONST and KEEP_BSS with
;	class 'FAR_DATA', all addressed with respect to KEEP_GROUP.  The
;	present implementation provides no other segments for resident data.

;	Note that the C compiler uses the COMM directive to declare
;	uninitialised data, leaving it to the linker to assign a position,
;	typically outside the executable's disk image and almost certainly
;	above the fence between resident and non-resident data.  It seems
;	impossible to account in advance for the linker's assignment and
;	therefore uninitialised data is not retained.

KEEP			SEGMENT WORD PUBLIC 'FAR_DATA'
KEEP			ENDS
KEEP_CONST		SEGMENT WORD PUBLIC 'FAR_DATA'
KEEP_CONST		ENDS
KEEP_BSS		SEGMENT WORD PUBLIC 'FAR_DATA'
KEEP_BSS		ENDS

KEEP_GROUP		GROUP	KEEP, KEEP_CONST, KEEP_BSS

;   Paragraph-aligned fence for non-resident data:

NOKEEP_BDATA		SEGMENT PARA COMMON 'FAR_DATA'
NOKEEP_BDATA		ENDS

;   All other segments in the program, including those comprising the
;   default data group DGROUP, will be placed here.

;   ************************************************************************

