
/*  ************************************************************************  *
 *				    shift.c				      *
 *  ************************************************************************  */

/*  This file contains functions which perform 32-bit shifts by operating
    separately on 16-bit components to avoid internal calls to non-resident
    library functions.	*/

#include    <standard.h>

/*  Functions are provided for shifting left and right.  Small shifts
    operate separately on the high and low words, but in addition, a piece
    of one word moves to the other.  Shifts of more than 16 bits necessarily
    obliterate one component.  */

DWORD _cdecl shl (DWORD x, WORD cl)
{
    WORD high, low, middle;

    if (cl < 16) {
	high = HIGHWORD (x) << cl;
	low = LOWWORD (x) << cl;
	middle = LOWWORD (x) >> 16 - cl;
	return (MK_DWORD (high | middle, low));
    }
    else return (MK_DWORD (LOWWORD (x) << cl - 16, 0));
}

DWORD _cdecl shr (DWORD x, WORD cl)
{
    WORD high, low, middle;

    if (cl < 16) {
	low = LOWWORD (x) >> cl;
	high = HIGHWORD (x) >> cl;
	middle = HIGHWORD (x) << 16 - cl;
	return (MK_DWORD (high, middle | low));
    }
    else return (MK_DWORD (0, HIGHWORD (x) >> cl - 16));
}

/*  ************************************************************************  */

