
/*  ************************************************************************  *
 *	     sft.c - a small collection of SFT management routines	      *
 *  ************************************************************************  */

#include    <sft.h>
#include    <standard.h>
#include    <sysvars.h>

static struct SFT _far *GetIndexedSFT (WORD);

struct SFT_HEADER _far *GetFirstSFTHeader (void)
{
    struct SYSVARS _far *sysvarsptr;

    _asm {					    // straightforward -
	    mov     ah,52h			    // just lift the
	    int     21h 			    // required address
	    mov     word ptr sysvarsptr,bx	    // from the SYSVARS
	    mov     word ptr sysvarsptr+2,es	    // structure in the
    }						    // DOS kernel
    return (sysvarsptr -> lpSFTs);
}

WORD GetSFTCount (void)
{
    struct SFT_HEADER _far *sftheader;
    WORD count;

    /*	Follow the chain of SFT blocks, noting the number of SFTs in each
	and accumulating a total.  */

    sftheader = GetFirstSFTHeader ();
    count = 0;
    do {
	count += sftheader -> wCount;
	sftheader = sftheader -> lpNext;
    } while ((WORD) sftheader != -1);
    return (count);
}

WORD GetSFTSize (void)
{
    struct SFT _far *sftptr, _far *sftmark = NULL;
    register WORD index, sftsize = 0;
    WORD count = GetSFTCount ();
    WORD diff;

    for (index = 0; index < count; index ++) {

	/*  Get the address of the next SFT.  Since the index is limited by
	    the count obtained before entering the loop, failure to obtain
	    an SFT address would be an exception situation that warrants
	    failing the function.  */

	sftptr = GetIndexedSFT (index);
	if (sftptr == NULL) return (0);

	/*  If this SFT and the one before it both lie in the same block,
	    then the difference of their addresses should be the SFT size.
	    Record this if it is not yet known, else verify consistency.  */

	if ((_segment) sftptr == (_segment) sftmark) {
	    diff = (BYTE _far *) sftptr - (BYTE _far *) sftmark;
	    if (sftsize == 0) sftsize = diff;
	    else if (sftsize != diff) return (0);
	}

	sftmark = sftptr;
    }
    return (sftsize);
}

struct SFT _far *GetIndexedSFT (WORD index)
{
    _asm {				// compiler preserves register di
	    mov     bx,index
	    mov     ax,1216h		// undocumented function takes index
	    int     2Fh 		// in BX and returns with es:di
	    jc	    fail		// addressing the corresponding SFT

	    mov     ax,di		// carry flag clear on success
	    mov     dx,es
	    jmp     done

	fail:				// carry flag set to indicate error
	    xor     ax,ax
	    cwd

	done:
    }
}

/*  ************************************************************************  */

