
/*  ************************************************************************  *
 *	     diskhook.c - demonstrates hooking int 13h below IO.SYS	      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <errno.h>
#include    <process.h>
#include    <signal.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <arena.h>
#include    <intpack.h>
#include    <standard.h>

char description [] = "\n\
Demonstrates how int 2Fh function 13h hooks disk services below IO.SYS.\n\
Use with the /W switch to also enable an interface for compatibility\n\
with the Windows 3.1 FastDisk feature.\n";

char instructions [] = "\n\
Int 13h hooked below IO.SYS.  To demonstrate the insertion, the handler\n\
prevents int 13h from writing to the first physical hard disk.\n\
\n\
A command processor is being started to allow investigation of this\n\
situation with whatever tools you prefer.  Type EXIT when satisfied.\n";

/*  Functions and data which may be needed to force the new command
    processor to load in conventional memory  */

extern WORD _dos_getstrategy (WORD *);		    // in memstrat.c
extern WORD _dos_getumblink (BOOL *);		    // in memstrat.c
extern WORD _dos_setstrategy (WORD);		    // in memstrat.c
extern WORD _dos_setumblink (BOOL);		    // in memstrat.c

WORD wOrgMemoryStrategy;
BOOL bOrgUMBLink;

/*  Function prototypes and data declarations for interrupt handling.  The
    program installs an int 13h routine below IO.SYS, but must also trap
    int 19h and may install an int 2Fh routine to resolve a conflict with
    the Windows 3.1 FastDisk feature if the /Windows switch is supplied.

    Use a typedef when it seems better to emphasise that a variable or
    function parameter is a pointer, rather than let this identity get lost
    amid all the keywords and brackets.

    The indisputably ugly prototype for the HookInt13hBelowIOSYS function
    allows all problems with interrupt windows to be dealt with inside that
    function and can greatly simplify the program's interrupt handlers.  */

typedef void _interrupt _far INTERRUPT_HANDLER (struct INTPACK);

void HookInt13hAtESBXAddress (INTERRUPT_HANDLER *, INTERRUPT_HANDLER **);

INTERRUPT_HANDLER *lpOrgESBXAddress = NULL, *lpCurrentESBXAddress;
INTERRUPT_HANDLER *lpOrgInt19h = NULL;
INTERRUPT_HANDLER *lpOrgInt2Fh = NULL;

void _interrupt _far Int13h (struct INTPACK);
void _interrupt _far Int19h (struct INTPACK);
void _interrupt _far Int2Fh (struct INTPACK);

#pragma intrinsic (_disable, _enable)

/*  Function for access to the Programmable Interrupt Controller, if
    electing to deal with interrupt windows by the brutal method of using
    hardware to inhibit interrupt generation  */

#ifdef _PIC_
WORD SetInterruptMask (WORD);
#endif

/*  Flag and function for Windows support  */

BOOL bWindows = FALSE;
BOOL SupportFastDisk (INTERRUPT_HANDLER *);

/*  Exit handling, including a message to cover an obscure circumstance in
    which proper cleanup is impossible.  */

int _far HardErrorTrap (void);
void RestoreInterrupts (void);
void RestoreMemoryUsage (void);

#ifndef _PIC_
char FatalInt19hMessage [] = "\n\
Int 19h warm boot attempted during window in which int 13h chain has been\n\
disconnected underneath IO.SYS and correct address for restoration is\n\
unknown.  System halted - reboot computer with Ctrl-Alt-Del or reset.";
#endif

void PutError (char *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register char *ptr, *pCOMSPEC;
    int exitcode;

    /*	Int 2Fh function 13h is certainly not available before DOS 3.10.  */

    if (MK_WORD (_osmajor, _osminor) < MK_WORD (3, 10)) {
	PutError ("requires DOS version 3.10 or higher");
	return (-1);
    }

    /*	Process the command line, looking for optional switches.  There
	follows a boring template approach to this generally boring task,
	which receives no further comment in the book.	*/

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;
	    if (ptr [0] == '?' AND ptr [1] == '\0') {
		printf (description);
		return (0);
	    }
	    if (strnicmp ("Windows", ptr, strlen (ptr)) == 0) {
		bWindows = TRUE;
		continue;
	    }
	    ptr --;
	    PutError ("invalid switch %s", ptr);
	    return (-1);
	}
	else {
	    PutError ("invalid parameter %s", ptr);
	    return (-1);
	}
    }

    /*	Locate the command processor.  */

    pCOMSPEC = getenv ("COMSPEC");
    if (pCOMSPEC == NULL) {
	PutError ("cannot find command processor for test");
	return (-1);
    }

    /*	Make sure that the program does not terminate without restoring the
	interrupt vector chains to their original condition.  That this
	program executes another would complicate any attempt to deal with
	Ctrl-C non-trivially (whatever the signal function's provisions),
	so just defeat it.  */

    atexit (RestoreInterrupts);     // covers normal exit
    signal (SIGINT, SIG_IGN);	    // prevents termination via int 23h
    _harderr (HardErrorTrap);	    // prevents termination via int 24h

    /*	By default, the WDCTRL virtual device driver needed for the FastDisk
	feature in Windows 3.1 Enhanced mode refuses to install if it finds
	an int 13h handler already sitting between IO.SYS and the BIOS.  An
	exception is offered through an int 2Fh service - to which this
	program aims to respond positively if executed with the /w switch.  */

    if (bWindows) {

	/*  Get the current value of the es:bx pointer in IO.SYS by
	    hooking into the chain and unhooking straight away.  */

	HookInt13hAtESBXAddress (Int13h, &lpCurrentESBXAddress);
	HookInt13hAtESBXAddress (lpCurrentESBXAddress, NULL);

	/*  Reproduce the WDCTRL query (at least tolerably well).  Cancel
	    the /w switch if WDCTRL's installation would be prevented by
	    an int 13h handler already closer to the BIOS than this
	    program's handler will be.  */

	if (NOT SupportFastDisk (lpCurrentESBXAddress)) {
	    printf ("\nWARNING:  cannot support Windows FastDisk\n");
	    bWindows = FALSE;
	}
    }

    /*	Hook int 19h before int 13h.  Inserting an int 13h handler between
	IO.SYS and the BIOS has the (possibly subtle) side-effect of
	defeating the IO.SYS mechanism for restoring int 13h to its BIOS
	default in the event of an int 19h warm boot.  Responsibility for
	maintaining system integrity lies entirely with the program that
	manipulates the int 13h chain at this low level.  This program's
	int 19h handler makes sure that changes to the int 13h chain (if
	known to have been made) are undone before the interrupt is passed
	any futher.  */

    lpOrgInt19h = _dos_getvect (0x19);
    _dos_setvect (0x19, Int19h);

    /*	Now hook int 13h, supplying the address of the variable in which
	the original es:bx address is expected by the int 13h and int 19h
	handlers.  Depending on the treatment of hardware interrupts during
	the use of int 2Fh function 13h, passing the address of the storage
	variable may allow the HookInt13hAtESBXAddress function to resolve
	internally a number of headaches caused by the way that the int 13h
	and int 19h handlers cannot act sensibly until they know the ds:dx
	and es:bx addresses.  */

    HookInt13hAtESBXAddress (Int13h, &lpOrgESBXAddress);

    /*	Just check that int 2Fh function 13h really seems to have done
	something - apparently, some OEMs do not implement the function
	even for DOS versions >= 3.10.	*/

    if (lpOrgESBXAddress == Int13h) {
	PutError ("int 2Fh function 13h apparently not supported");
	return (-1);
    }

    /*	If co-operation with Windows is desired, hook int 2Fh.	*/

    if (bWindows) {
	lpOrgInt2Fh = _dos_getvect (0x2F);
	_dos_setvect (0x2F, Int2Fh);
    }

    /*	Before running the command processor for the test, remember to force
	the use of conventional memory.  If the upper memory link is enabled
	and a loadhigh strategy set when this program starts, the new
	command processor and all its children will be loaded into upper
	memory (unless the supply is exhausted).  This would limit the
	testing environment (for instance, Windows enhanced mode will not
	start if WIN386 has been loaded into upper memory, or indeed, if it
	has been loaded into conventional memory while the loadhigh strategy
	is in force).  */

    if (_osmajor >= 5) {
	_dos_getstrategy (&wOrgMemoryStrategy);
	_dos_getumblink (&bOrgUMBLink);
	if (wOrgMemoryStrategy != FIRST_FIT OR bOrgUMBLink) {
	    atexit (RestoreMemoryUsage);
	    _dos_setstrategy (FIRST_FIT);
	    _dos_setumblink (OFF);
	}
    }

    /*	Now run the command processor, as described in the instructions.  */

    printf (instructions);
    exitcode = spawnl (P_WAIT, pCOMSPEC, pCOMSPEC, NULL);
    if (exitcode == -1) {
	PutError ("%s", strerror (errno));
	return (-1);
    }
    else {
	printf ("\nDISKHOOK:  Int 13h hooking test completed.\n");
	return (exitcode);
    }
}

void HookInt13hAtESBXAddress (INTERRUPT_HANDLER *newhandler,
				INTERRUPT_HANDLER **oldhandlerstorage)
{
    INTERRUPT_HANDLER *dsdxaddress;
    INTERRUPT_HANDLER *esbxaddress;

    /*	One way to deal with some design problems of int 2Fh function 13h
	is simply to prevent hardware interrupts from occurring while
	hooking the int 13h chain (note however, that this does still leave
	a problem with the non-maskable interrupt).  */

    #ifdef _PIC_				// all bits set in new mask
    WORD oldmask = SetInterruptMask (0xFFFF);	// to inhibit all hardware
    #endif					// interrupts

    /*	The intention is to thread the new handler into the int 13h chain as
	close as possible to the ROM BIOS.  Unfortunately, key points in the
	chain are not known in advance of calling int 2Fh function 13h.
	Begin therefore, by installing the new handler as the only int 13h
	handler acting underneath IO.SYS, then rearrange the chain once more
	information is available.  */

    _asm {
	    push    ds				// must save ds=DGROUP

	    lds     dx,newhandler		// if interrupts can occur,
	    les     bx,newhandler		// then newhandler must be
	    mov     ah,13h			// able to replace BIOS for
	    int     2Fh 			// a while

	    mov     word ptr dsdxaddress,dx	// ds:dx and es:bx returned
	    mov     word ptr dsdxaddress+2,ds	// by int 2Fh function 13h
	    mov     word ptr esbxaddress,bx	// may both be needed for
	    mov     word ptr esbxaddress+2,es	// chain reconnection

	    pop     ds				// restore ds=DGROUP
    }

    /*	This blind use of the interface leaves the int 13h chain is in as
	many as three disconnected pieces:

	    IO.SYS directly to newhandler
	    dsdxaddress  --> possibly other handlers --> newhandler
	    esbxaddress  --> possibly other handlers --> ROM BIOS

	Unless the ds:dx and es:bx addresses are equal (in which case, the
	second piece effectively does not exist), the flow from IO.SYS must
	be re-routed to the ds:dx address.  */

    if (dsdxaddress != esbxaddress) {
	_asm {
		push	ds
		lds	dx,dsdxaddress
		les	bx,newhandler
		mov	ah,13h
		int	2Fh
		pop	ds
	}
    }

    /*	Of course, the main reconnection is to link the new handler with the
	es:bx address.	Unless hardware interrupts are inhibited, the moment
	that the new handler possesses the value currently in the local
	variable esbxaddress, int 13h requests will be able to reach the
	BIOS.  Int 13h will function as before, with just the one addition
	of the new handler in front of the previous es:bx address.  Note the
	importance of taking this step and the preceding one in the right
	order;	to do otherwise would risk an int 13h that skips the
	handlers between the original ds:dx address and the new handler.  */

    if (oldhandlerstorage != NULL) {
	#ifndef _PIC_				// remember to disable
	_disable ();				// interrupts, lest an
	#endif					// int 13h occur while the
	*oldhandlerstorage = esbxaddress;	// pointer used for indirect
	#ifndef _PIC_				// transfer is only
	_enable ();				// half-saved
	#endif
    }

    /*	If however, hardware interruption is defeated, then to some extent,
	the ordering discussed above is immaterial - just as long as
	everything is in place by the time that the interrupt mask is
	restored.  */

    #ifdef _PIC_
    SetInterruptMask (oldmask);
    #endif
}

#ifdef _PIC_
WORD SetInterruptMask (WORD mask)
{
    _asm {
	    push    dx
	    pushf
	    cli
	    in	    al,21h	    // read PIC1 Interrupt Mask Register
	    mov     dl,al	    //	 and save in dl
	    in	    al,0A1h	    // read PIC2 Interrupt Mask Register
	    mov     dh,al	    //	 and save in dh
	    mov     ax,mask	    // write new masks to each PIC
	    out     21h,al
	    mov     al,ah
	    out     0A1h,al
	    mov     ax,dx	    // retrieve original mask into ax
	    popf		    //	 to return as value of C function
	    pop     dx
    }
}
#endif

void _interrupt _far Int13h (struct INTPACK r)
{
    _enable ();

    /*	Given the way that int 2Fh function 13h works, the address to which
	this trial handler transfers uninteresting requests will not be
	known until the handler has already been threaded into the interrupt
	chain.	Prevent the handler from doing anything non-trivial until
	the old address has been recorded safely.  Of course, nothing need
	be done about this case if its occurrence is prevented by using the
	PIC to inhibit interrupts during the critical moments.	*/

    #ifndef _PIC_
    if (lpOrgESBXAddress == NULL) {
	r.ax = 0x0100;
	r.flags |= 0x0201;
	*((BYTE _far *) MK_FP (0x0040, r.dl & 0x80 ? 0x0074 : 0x0041)) = 0x01;
	return;
    }
    #endif

    /*	For demonstration purposes, this handler simply defeats all attempts
	to write to the first physical hard disk.  Everything else is passed
	on to the true BIOS int 13h handler (or to some interloper such as a
	boot sector virus that hooked int 13h before IO.SYS).  */

    if (r.ah == 0x03 AND r.dl == 0x80) {
	r.ax = 0x0100;
	r.flags |= 0x0201;
	*((BYTE _far *) MK_FP (0x0040, r.dl & 0x80 ? 0x0074 : 0x0041)) = 0x01;
	return;
    }
    else _chain_intr (lpOrgESBXAddress);
}

void _interrupt _far Int19h (struct INTPACK r)
{
    _enable ();

    /*	If interrupts are allowed during the disk hook operations (i.e., not
	prevented via the PIC), then it is possible (though admittedly,
	exceptional) for int 19h to be issued while the address to which
	int 13h must be restored is not yet known.  In this case, int 19h
	cannot be allowed to proceed.  Make this program the current process
	for safety while writing an error message (there is a strong
	argument in favour of using the video BIOS functions for this sort
	of message, particularly to ensure that the screen is clear), then
	die.  */

    #ifndef _PIC_
    if (lpOrgESBXAddress == NULL) {
	_asm {
		mov	bx,_psp
		mov	ah,50h
		int	21h
	}
	PutError (FatalInt19hMessage);
	_asm {
	    die:
		jmp	die
	}
    }
    #endif

    RestoreInterrupts ();
    _chain_intr (lpOrgInt19h);
}

void _interrupt _far Int2Fh (struct INTPACK r)
{
    _enable ();

    /*	If the WDCTRL virtual device decides that es:bx returned by int 2Fh
	function 13h does not address an int 13h entry point in ROM, it
	issues a BLOCKDEV Device Broadcast to determine whether or not its
	particular Windows FastDisk service may be installed.  In this
	situation, register AX contains 1607h and register BX contains the
	device ID 0010h for BLOCKDEV (the device which actually hooks
	int 13h between IO.SYS and the BIOS).  The service in question is
	number 0003h, placed in register CX.  If cx is not cleared to zero,
	WDCTRL aborts its installation.  */

    if (r.ax == 0x1607 AND r.bx == 0x0010 AND r.cx == 0x0003) {

	/*  Before returning zero in cx to let WDCTRL install, check that
	    no int 13h handlers have been installed through the es:bx
	    pointer since this program's int 13h handler was added.  */

	HookInt13hAtESBXAddress (lpOrgESBXAddress, &lpCurrentESBXAddress);
	HookInt13hAtESBXAddress (lpCurrentESBXAddress, NULL);
	if (lpCurrentESBXAddress == Int13h) r.cx = 0;
    }
    else _chain_intr (lpOrgInt2Fh);
}

BOOL SupportFastDisk (INTERRUPT_HANDLER *current_esbx)
{
    /*	This is a reproduction of WDCTRL's own method for deciding whether
	or not to install an int 13h handler through the es:bx pointer, with
	the omission of the test for Quarterdeck's Stealth.  */

    if ((_segment) current_esbx >= 0xA000) return (TRUE);

    _asm {
	    mov     cx,0003h
	    mov     bx,0010h
	    mov     ax,1607h
	    int     2Fh
	    mov     ax,0001h	    ; TRUE if cx returns as zero
	    jcxz    done
	    dec     ax		    ; FALSE otherwise
	done:
    }
}

void RestoreMemoryUsage (void)
{
    _dos_setstrategy (wOrgMemoryStrategy);
    _dos_setumblink (bOrgUMBLink);
}

void RestoreInterrupts (void)
{
    if (lpOrgESBXAddress != NULL)
	HookInt13hAtESBXAddress (lpOrgESBXAddress, NULL);
    if (lpOrgInt19h != NULL) _dos_setvect (0x19, lpOrgInt19h);
    if (lpOrgInt2Fh != NULL) _dos_setvect (0x2F, lpOrgInt2Fh);
}

int _far HardErrorTrap (void)
{
    return (_HARDERR_FAIL);
}

void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nDISKHOOK error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

