
/*  ************************************************************************  *
 *				   getint13.c				      *
 *  ************************************************************************  */

/*  This file shows how to use the undocumented int 2Fh function 13h to
    locate the entry point to the BIOS int 13h handler (or some replacement
    installed already).  */

#include    <intpack.h>     // supplies PUSHA structure, etc, for access to
			    // caller's registers in functions declared with
			    // the _interrupt keyword in Microsoft C -
			    // may have to be replaced for other compilers
#include    <standard.h>

static void _interrupt _far DummyInt13h (struct INTPACK);

void (_interrupt _far *GetBIOSInt13h (void)) (void)
{
    _asm {
	    push    ds	    // preserve ds throughout function (it is
			    // expected to change) - pessimists may want
			    // to preserve registers bp, si and di also

    /*	Call int 2Fh function 13h to install a dummy replacement for both
	the BIOS int 13h handler and the IO.SYS machine-specific filter.
	The function expects new addresses for these two routines in es:bx
	and ds:dx respectively, returning the previous addresses in these
	same registers.  Thus, it is not possible to obtain the addresses
	without disturbing the int 13h chain.  */

	    push    cs
	    pop     ds
	    mov     dx,offset DummyInt13h
	    push    cs
	    pop     es
	    mov     bx,dx
	    mov     ah,13h
	    int     2Fh

	    push    es	    // save es:bx on stack, to return as value of
	    push    bx	    // C function

    /*	Call int 2Fh function 13h a second time to reinstall the routines
	that were temporarily replaced by the dummy handler.  */

	    mov     ah,13h
	    int     2Fh

	    pop     ax	    // retrieve address of BIOS int 13h handler
	    pop     dx	    // into dx:ax as return value of C function

	    pop     ds	    // restore ds = DGROUP
    }
}

static void _interrupt _far DummyInt13h (struct INTPACK r)
{
    /*	Fail the int 13h call as if for an unsupported function.  */

    r.ax = 0x0100;	    // return with ah = 0x01 as error code

    r.flags |= 0x0201;	    // return with interrupts enabled (0x0200),
			    // carry flag set (0x0001) to indicate error,
			    // all other flags unchanged

    /*	For completeness, also record the simulated error code in the BIOS
	data variable (0040:0041 for floppy disks, 0040:0074 for hard disks)
	consulted by int 13h function 01h.  */

    *((BYTE _far *) MK_FP (0x0040, r.dl & 0x80 ? 0x0074 : 0x0041)) = 0x01;
}

/*  There follows a small test rig.  */

#include    <stdio.h>
#include    <stdlib.h>

void main (void)
{
    if (MK_WORD (_osmajor, _osminor) < MK_WORD (3, 10)) return;
    printf ("\nIO.SYS thinks that the BIOS int 13h handler is at %Fp\n",
	GetBIOSInt13h ());
}

/*  ************************************************************************  */

