
/*  ************************************************************************  *
 *				    malloc.c				      *
 *  ************************************************************************  */

/*  This file provides a generalisation of DOS memory allocation to allow
    selection of a memory block by location as well as size.  */

#include    <dos.h>
#include    <stdlib.h>

#include    <standard.h>

extern WORD MaskMemoryRegion (_segment, _segment, WORD *);  // memblock.c
extern WORD ReleaseMask (WORD *);			    // memblock.c

extern WORD _dos_getumblink (BOOL *);			    // memstrat.c
extern WORD _dos_setumblink (BOOL);			    // memstrat.c

extern void _dos_getbreakflag (BYTE *); 		    // break.c
extern void _dos_setbreakflag (BYTE);			    // break.c

WORD _dos_special_allocmem (WORD size, _segment base, _segment ceiling,
				    _segment *baseresult, WORD *sizeresult)
{
    WORD handle;
    register WORD errorcode;
    BYTE oldbreakflag;
    BOOL oldumblink;
    WORD temp;

    if (_osmajor >= 5) {
	_dos_getumblink (&oldumblink);
	if (NOT oldumblink) {
	    _dos_getbreakflag (&oldbreakflag);
	    if (oldbreakflag) _dos_setbreakflag (OFF);
	    _dos_setumblink (ON);
	}
    }

    /*	Enforce 0x0000 <= base <= base + size <= ceiling <= 0xFFFF for the
	block to allocate.  */

    if (0xFFFF - base < size) size = 0xFFFF - base;
    if (ceiling < base) ceiling = base + size;
    else if (ceiling - base < size) size = ceiling - base;

    errorcode = MaskMemoryRegion (base, ceiling, &handle);

    if (errorcode == 0) {
	errorcode = _dos_allocmem (size, &temp);
	if (errorcode == 0) {
	    *baseresult = temp;
	    *sizeresult = size;
	}
	else {
	    *sizeresult = temp;
	}
    }

    if (_osmajor >= 5) {
	if (NOT oldumblink) _dos_setumblink (oldumblink);
	if (oldbreakflag) _dos_setbreakflag (oldbreakflag);
    }

    ReleaseMask (&handle);

    return (errorcode);
}

/*  ************************************************************************  */

