
/*  ************************************************************************  *
 *				   memblock.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <malloc.h>
#include    <stdlib.h>

#include    <doserror.h>
#include    <standard.h>

struct MEMBLOCK {
    struct MEMBLOCK *pNext;	    // for list management only
    _segment spBase;
    _segment spCeiling;
};

WORD MaskMemoryRegion (_segment, _segment, struct MEMBLOCK **);
void ReleaseMask (struct MEMBLOCK **);

static struct MEMBLOCK *AddBlock (_segment, _segment);
static void RemoveBlock (struct MEMBLOCK *);

static struct MEMBLOCK *pLowestBlock;

static WORD AddLargestBlock (void);
static WORD PunchHole (_segment, _segment);

WORD MaskMemoryRegion (_segment base, _segment ceiling,
						struct MEMBLOCK **handleptr)
{
    register WORD errorcode;

    /*	If the range is trivial, return immediately.  */

    if (base == 0x0000 AND ceiling == 0xFFFF) {
	*handleptr = NULL;
	return (0);
    }

    pLowestBlock = NULL;
    do {
	errorcode = AddLargestBlock ();
    } while (errorcode == 0);

    if (errorcode == E_OUT_OF_MEMORY) errorcode = PunchHole (base, ceiling);

    if (errorcode != 0) ReleaseMask (&pLowestBlock);
    *handleptr = pLowestBlock;
    return (errorcode);
}

static WORD PunchHole (_segment base, _segment ceiling)
{
    register struct MEMBLOCK *ptr;
    register WORD errorcode = 0;
    WORD size;
    struct MEMBLOCK dummy;

    for (ptr = pLowestBlock; ptr != NULL; ptr = ptr -> pNext) {

	/*  Skip blocks that lie completely outside the "hole".  Since
	    the list of MEMBLOCK structures is sorted by increasing base
	    address, the search may be abandoned once the base address
	    exceeds the upper limit of the desired region.  */

	if (ptr -> spCeiling < base) continue;
	if (ptr -> spBase > ceiling) break;

	if (ptr -> spBase < base OR ptr -> spCeiling > ceiling) {

	    if (ptr -> spBase < base) size = base - ptr -> spBase - 1;
	    else size = ceiling - ptr -> spBase;

	    errorcode = _dos_setblock (size, ptr -> spBase, &size);
	    if (errorcode != 0) break;

	    /*	Adjust the corresponding MEMBLOCK structure.  */

	    ptr -> spCeiling = ptr -> spBase + size;

	    errorcode = AddLargestBlock ();
	    if (errorcode != 0) break;

	    if (ptr -> spBase < base) continue;
	}

	errorcode = _dos_freemem (ptr -> spBase);
	if (errorcode != 0) break;

	dummy.pNext = ptr -> pNext;
	RemoveBlock (ptr);
	ptr = &dummy;
    }

    return (errorcode);
}

static struct MEMBLOCK *AddBlock (_segment base, _segment ceiling)
{
    register struct MEMBLOCK **mark, *ptr;
    struct MEMBLOCK *new;

    new = malloc (sizeof (struct MEMBLOCK));
    if (new == NULL) return (NULL);

    mark = &pLowestBlock;
    ptr = pLowestBlock;
    while (ptr != NULL AND ptr -> spBase <= base) {
	mark = &(ptr -> pNext);
	ptr = ptr -> pNext;
    }

    new -> pNext = ptr;
    new -> spBase = base;
    new -> spCeiling = ceiling;
    *mark = new;

    return (new);
}

static void RemoveBlock (struct MEMBLOCK *memblock)
{
    register struct MEMBLOCK **mark, *ptr;

    mark = &pLowestBlock;
    ptr = pLowestBlock;
    while (ptr != NULL AND ptr != memblock) {
	mark = &(ptr -> pNext);
	ptr = ptr -> pNext;
    }
    if (ptr == memblock) {
	*mark = ptr -> pNext;
	free (ptr);
    }
}

static WORD AddLargestBlock (void)
{
    register struct MEMBLOCK *memblock;
    register WORD errorcode;
    WORD base;
    WORD size;

    /*	Obtain the largest free block available from DOS.  */

    errorcode = _dos_allocmem (0xFFFF, &size);
    if (errorcode == 0) {
	base = size;
	size = 0xFFFF;
    }
    else if (errorcode == E_OUT_OF_MEMORY)
	errorcode = _dos_allocmem (size, &base);

    if (errorcode == 0) {
	memblock = AddBlock (base, base + size);
	if (memblock == NULL) errorcode = E_ACCESS_DENIED;
    }
    return (errorcode);
}

void ReleaseMask (register struct MEMBLOCK **starter)
{
    register struct MEMBLOCK *ptr;

    ptr = *starter;
    while (ptr != NULL) {
	*starter = ptr -> pNext;
	_dos_freemem (ptr -> spBase);
	free (ptr);
	ptr = *starter;
    }
}

/*  ************************************************************************  */

