
/*  ************************************************************************  *
 *		 spotload.c - top-level module for SPOTLOAD.EXE 	      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <process.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <arena.h>
#include    <standard.h>

char description [] = "\n\
Loads a given program into a specified memory region\n\
\n\
  SPOTLOAD [/B:base [/C:ceiling]] [options] program [arguments]\n\
\n\
Base and ceiling segment addresses are expected in hexadecimal form.\n\
Use /R to restrict the child program to this same area of memory.\n\
\n\
Other options set the memory allocation environment that will apply\n\
while the child program executes:-\n\
\n\
  /U:[OFF|ON]		enable or disable the upper memory link\n\
  /S:[FIRST|BEST|LAST]	set a bare memory allocation strategy\n\
  /L:[OFF|ON]		set a preference for upper memory (loadhigh)\n\
  /H:[OFF|ON]		set an insistence on upper memory (highonly)\n\
";

/*  Variables to configure the child program's memory operations during its
    loading and executing stages  */

extern _segment spBase, spCeiling;		    // while loading

extern WORD wChildsStrategy;			    // remainder apply
extern BOOL bChildsUMBLink;			    // while child executes
extern BOOL bSetChildsStrategy, bSetChildsUMBLink;

extern BOOL bRestrictChildsMemory;

/*  Functions for strategy control  */

extern WORD _dos_getstrategy (WORD *);		    // in memstrat.c
extern WORD _dos_getumblink (BOOL *);		    // in memstrat.c
extern WORD _dos_setstrategy (WORD);		    // in memstrat.c
extern WORD _dos_setumblink (BOOL);		    // in memstrat.c

static BOOL ReadStrategy (char *, WORD *);
static char *strategies [] = {"first", "best", "last", NULL};

/*  Miscellany	*/

extern BOOL ReadBoolean (char *, BOOL *);
extern BOOL ReadHexWord (char *, WORD *);

void PutError (char *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register CHAR *ptr;
    register WORD ch;
    WORD temp;
    int exitcode;

    /*	Reject early DOS versions.  */

    if (MK_WORD (_osmajor, _osminor) < MK_WORD (3, 10)) {
	PutError ("requires DOS version 3.10 or higher");
	return (-1);
    }

    /*	Obtain (run-time) default values for the child program's memory
	environment.  */

    if (_osmajor >= 5) {
	_dos_getstrategy (&wChildsStrategy);
	_dos_getumblink (&bChildsUMBLink);
    }

    /*	Process the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ch = ptr [1] | (BYTE) 0x20;
	    if (ptr [2] == '\0') {
		if (ch == '?') {
		    printf (description);
		    return (0);
		}
		else if (ch == 'r') bRestrictChildsMemory = TRUE;
	    }
	    else if (ptr [2] == ':') {
		if (ch == 'b' OR ch == 'c') {
		    if (NOT ReadHexWord (ptr + 3, &temp)) {
			PutError ("bad number in switch %s", ptr);
			return (-1);
		    }
		    if (ch == 'b') spBase = temp;
		    else spCeiling = temp;		    // if (ch == 'c')
		}
		else if (ch == 'h' OR ch == 'l' OR ch == 'u') {
		    if (NOT ReadBoolean (ptr + 3, &temp)) {
			PutError ("bad Boolean in switch %s", ptr);
			return (-1);
		    }
		    if (_osmajor >= 5) {
			if (ch == 'h') {
			    if (temp) temp = HIGHONLY;
			    temp = (wChildsStrategy & HIGHONLY) ^ temp;
			    if (temp) {
				wChildsStrategy ^= temp;
				bSetChildsStrategy = TRUE;
			    }
			}
			else if (ch == 'l') {
			    if (temp) temp = LOADHIGH;
			    temp = (wChildsStrategy & LOADHIGH) ^ temp;
			    if (temp) {
				wChildsStrategy ^= temp;
				bSetChildsStrategy = TRUE;
			    }
			}
			else {				    // if (ch == 'u')
			    if (bChildsUMBLink ^ temp) {
				bChildsUMBLink = temp;
				bSetChildsUMBLink = TRUE;
			    }
			}
		    }
		}
		else if (ch == 's') {
		    if (NOT ReadStrategy (ptr + 3, &temp)) {
			PutError ("bad strategy in switch %s", ptr);
			return (-1);
		    }
		    if (_osmajor >= 5) {
			temp = (wChildsStrategy & STRATEGY_MASK) ^ temp;
			if (temp) {
			    wChildsStrategy ^= temp;
			    bSetChildsStrategy = TRUE;
			}
		    }
		}
	    }
	    else {
		PutError ("invalid switch %s", ptr);
		return (-1);
	    }
	}
	else break;
    }
    if (argc == 0) {
	PutError ("requires name of program to load");
	return (-1);
    }

    exitcode = spawnvp (P_WAIT, *argv, argv);

    if (exitcode == -1) PutError ("%s", strerror (errno));
    return (exitcode);
}

static BOOL ReadStrategy (char *ptr, WORD *result)
{
    register char **mark;
    char *str;
    WORD temp;

    mark = strategies;
    temp = FIRST_FIT;
    while (str = *mark ++, str != NULL) {
	if (stricmp (str, ptr) == 0) {
	    *result = temp;
	    return (TRUE);
	}
	temp ++;		// successive strategies differ by 1
    }
    return (FALSE);
}

void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nSPOTLOAD fatal error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

