
;   ************************************************************************
;   *				  setlimit.asm				   *
;   ************************************************************************

;   This file presents a minimal routine that steps up to protected mode
;   temporarily just to load the invisible part of the ES register with some
;   given limit.  

.MODEL			SMALL, C
.386P

SEGMENT_DESCRIPTOR	STRUCT
  limit 			dw	?
  base			dw	?
  base_middle_8 	db	?
  access_rights 	db	?
  extra_byte		db	?
  base_high_8		db	?
SEGMENT_DESCRIPTOR	ENDS

TABLE_DESCRIPTOR	STRUCT
  limit 			dw	?
  base			dd	?
TABLE_DESCRIPTOR	ENDS

;   ========================================================================

.DATA

;   The Global Descriptor Table used by the routine below provides just one
;   selector other than the null selector.  Interrupts are disabled while
;   the procedure runs in protected mode and in the absence of processor
;   exceptions (or non-maskable interrupts, which the caller must disable),
;   no need exists for an Interrupt Descriptor Table, code segment or stack
;   segment, but any serious application ought to provide a more
;   substantial, secure environment.

gdt			SEGMENT_DESCRIPTOR	<>, 
						< , 0000h, 00h, 93h>

ES_selector		=	SIZEOF SEGMENT_DESCRIPTOR
ES_descriptor		=	gdt + ES_selector

fpGDT			TABLE_DESCRIPTOR	<SIZEOF gdt - 0001h>

;   ========================================================================

.CODE

SetESLimit		PROC	PUBLIC USES es, limit:DWORD

;   Set the limit fields in the descriptor for the test segment.  Note that
;   limits requiring more than the 20 bits available in the descriptor must
;   be stored as some multiple of the ubiquitous 4KB page.  The granularity
;   bit (the MSB in the byte that also contains the four highest bits of the
;   limit) must be set to indicate that the limit has been scaled to fit.

			xor	bl,bl
			mov	eax,[limit]
			test	eax,0FFF00000h
			jz	@f
			shr	eax,12
			mov	bl,80h

@@:
			mov	[ES_descriptor.limit],ax
			shr	eax,10h
			or	al,bl
			mov	[ES_descriptor.extra_byte],al

;   Calculate the base address of the GDT and load the corresponding table
;   register.  Make sure that interrupts are in fact disabled.

			mov	ax,ds
			movzx	eax,ax
			shl	eax,4
			add	eax,OFFSET gdt
			mov	[fpGDT.base],eax

			mov	bx,ES_selector

			pushf
			cli

			lgdt	fword ptr [fpGDT]

;   Enter protected mode to load the ES register with the selector to the
;   prepared segment.  Flush the instruction queue to ensure that the
;   descriptor cache for ES is not loaded while the processor is still in
;   real mode.

			mov	eax,cr0
			or	al,01h
			mov	cr0,eax
			jmp	@f
@@:
			mov	es,bx

;   Return to real mode and restore the caller's interrupt state.

			and	al,NOT 01h
			mov	cr0,eax

			popf

;   Epilogue code restores the visible part of the ES register.  Flush the
;   instruction queue beforehand to make sure that the processor is back in
;   real mode when it decodes the POP ES instruction.  The processor
;   recalculates the base address, but does it also impose a 16-bit segment
;   size of 64KB?

			jmp	@f
@@:
			ret

SetESLimit		ENDP

;   ************************************************************************

END

