
;   ************************************************************************
;   *				  getlimit.asm				   *
;   ************************************************************************

;   This file contains a function that returns the 32-bit segment limit
;   currently applied to the ES register.

.MODEL			SMALL, C
.386

;   ========================================================================

.CODE

GetESLimit		PROC	PUBLIC USES ds

;   Install handlers for the real-mode Invalid Opcode and General Protection
;   faults.  Store the current vector addresses in the code segment for
;   simplicity.  Disable interrupts throughout the test, but note that the
;   caller must do this beforehand if the validity of the result on return
;   is to be ensured.

			xor	ax,ax
			mov	ds,ax

			mov	ax,cs
			shl	eax,10h
			mov	ebx,eax

			mov	ax,OFFSET int06h
			mov	bx,OFFSET int0Dh

			pushf
			cli

			xchg	eax,ds:[06h * 4]
			xchg	ebx,ds:[0Dh * 4]

			mov	cs:[lpOrgInt06h],eax
			mov	cs:[lpOrgInt0Dh],ebx

;   Try reading a byte of memory to see if a processor fault is generated.
;   Adopt a bisection algorithm to determine the highest address that does
;   not trigger a fault:

;   ebx = least linear address known to be invalid (or 00000000h)
;   ecx = current testing address
;   edx = greatest linear address known to be valid

;   Thus in general, edx <= ecx < ebx.

			xor	ebx,ebx
			xor	edx,edx
			xor	ecx,ecx
			dec	ecx

read::
			mov	al,es:[ecx]

;   If the address in ecx is valid (i.e., the previous instruction executes
;   normally), copy it to edx (highest known valid address) and continue.

			mov	edx,ecx
			jmp	@f

;   Attempting to read from an address regarded as invalid for segment ES
;   will generate an exception.  The handler (given below) will redirect
;   control here, so that ebx (lowest invalid address) may be updated
;   instead of edx.

fault::
			mov	ebx,ecx

@@:

;   Bisect the area of unknown behaviour and re-iterate.

			mov	ecx,ebx
			sub	ecx,edx
			jb	done
			shr	ecx,1
			jz	done
			add	ecx,edx
			jmp	read

done:

;   Restore the original exception handlers and the caller's interrupt 
;   state.
			mov	eax,cs:[lpOrgInt06h]
			mov	ds:[06h * 4],eax
			mov	eax,cs:[lpOrgInt0Dh]
			mov	ds:[0Dh * 4],eax

			popf

;   Return the 32-bit limit (now in edx) as dx:ax for the convenience of
;   C language callers.

			mov	ax,dx
			shr	edx,10h
			ret

GetESLimit		ENDP

;   ------------------------------------------------------------------------

int0Dh			PROC	FAR PRIVATE

;   Check that the handler has not been invoked while a hardware interrupt
;   at the conflicting level (IRQ5) is pending.

			push	ax
			mov	al,0Bh		;  signal the PIC to return
			out	20h,al		;  its In-Service register
			jmp	@f		;  on the next read
@@:
			in	al,20h
			test	al,01h SHL (0Dh - 08h)
			pop	ax
			jnz	chain

;   If the faulting instruction is the one expected to trigger a fault,
;   redirect control as described above.

			call	RedirectFault
			jnz	chain
			iret
chain:
			jmp	cs:[lpOrgInt0Dh]

int0Dh			ENDP

;   ------------------------------------------------------------------------

int06h			PROC	FAR PRIVATE

;   An Invalid Opcode handler is needed if running under a virtual machine
;   created by EMM386.	It is the same as the int 0Dh handler above, but
;   without needing to resolve a possible conflict with IRQ5.

			call	RedirectFault
			jnz	chain
			iret
chain:
			jmp	cs:[lpOrgInt06h]

int06h			ENDP

;   ------------------------------------------------------------------------

RedirectFault		PROC	PRIVATE USES ax, retaddr:DWORD

;   If the faulting instruction is the one expected to trigger a fault,
;   change the interrupt handler's return address and return zf = ZR.

			mov	ax,cs
			cmp	ax,word ptr [retaddr + 02h]
			jnz	@f
			cmp	word ptr [retaddr],OFFSET read
			jnz	@f

			mov	word ptr [retaddr],OFFSET fault

@@:
			ret

RedirectFault		ENDP

;   ------------------------------------------------------------------------

;   Storage for the int 06h and 0Dh address:

lpOrgInt06h		dd	?
lpOrgInt0Dh		dd	?

;   ************************************************************************

END

