
;   ************************************************************************
;   *				  setlimit.asm				   *
;   ************************************************************************

.MODEL			SMALL, C
.386P

;   Descriptor cache format for the 80386 machine state structure:

CACHE			STRUCT
  unknown_00h		db	?
  access_rights 	db	?
  unknown_02h		db	?
  unknown_03h		db	?
  base			dd	?
  limit 		dd	?
CACHE			ENDS

;   ========================================================================

.DATA

;   Statically allocated machine state structure for the 80386 LOADALL
;   instruction - observe that all segments are given access rights of 92h:

loadall_table		LABEL	BYTE

loadall_cr0		dd	?
loadall_eflags		dd	?
loadall_eip		dd	OFFSET loaded
loadall_edi		dd	?
loadall_esi		dd	?
loadall_ebp		dd	?
loadall_esp		dd	?
loadall_ebx		dd	?
loadall_edx		dd	?
loadall_ecx		dd	?
loadall_eax		dd	?
loadall_dr6		dd	?
loadall_dr7		dd	?
loadall_tr		dd	?
loadall_ldtr		dd	?
loadall_gs		dd	?
loadall_fs		dd	?
loadall_ds		dd	?
loadall_ss		dd	?
loadall_cs		dd	?
loadall_es		dd	?
loadall_tr_cache	CACHE	<>
loadall_idtr_cache	CACHE	<>
loadall_gdtr_cache	CACHE	<>
loadall_ldtr_cache	CACHE	<>
loadall_gs_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
loadall_fs_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
loadall_ds_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
loadall_ss_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
loadall_cs_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
loadall_es_cache	CACHE	<00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>

;   ========================================================================

.CODE

SetESLimit		PROC	PUBLIC USES gs fs es ds edi esi, limit:DWORD
			LOCAL	scratch:FWORD

;   Preserve registers important to system control (and whose values can be
;   determined in real mode) - the GDTR and IDTR table registers, CR0 and
;   the two debugging registers DR6 and DR7.

			sgdt	[scratch]
			mov	ax,word ptr [scratch]
			mov	word ptr [loadall_gdtr_cache.limit],ax
			mov	eax,dword ptr [scratch + 02h]
			mov	[loadall_gdtr_cache.base],eax

			sidt	[scratch]
			mov	ax,word ptr [scratch]
			mov	word ptr [loadall_idtr_cache.limit],ax
			mov	eax,dword ptr [scratch + 02h]
			mov	[loadall_idtr_cache.base],eax

			mov	eax,cr0
			mov	[loadall_cr0],eax
			mov	eax,dr6
			mov	[loadall_dr6],eax
			mov	eax,dr7
			mov	[loadall_dr7],eax

			pushfd
			pop	[loadall_eflags]

;   Make sure that all registers needed for access to the stack will retain
;   their values.  Visible and invisible components of the SS register must
;   be prepared - note that a 64KB limit for SS has been set statically.

			mov	[loadall_ebp],ebp
			mov	[loadall_esp],esp

			mov	ax,ss
			mov	word ptr [loadall_ss],ax
			movzx	eax,ax
			shl	eax,4
			mov	[loadall_ss_cache.base],eax

;   Similarly, make sure that CS:IP will address the instruction immediately
;   after LOADALL.  The value for the IP register is set statically.

			mov	ax,cs
			mov	word ptr [loadall_cs],ax
			movzx	eax,ax
			shl	eax,4
			mov	[loadall_cs_cache.base],eax

;   Assign the designated segment limit to the ES descriptor cache.

			mov	eax,[limit]
			mov	[loadall_es_cache.limit],eax

;   The LOADALL instruction expects the machine state structure at ES:DI.
;   Immediately after the LOADALL instruction, all registers (including the
;   invisible components of segment registers) have values taken from the
;   machine state structure.

			push	ds
			pop	es
			mov	di,OFFSET loadall_table

			db	0Fh, 07h
loaded::

;   Epilogue code restores the visible components of the segment registers,
;   forcing a recalculation of base addresses in the descriptor cache, but
;   leaving the new limit for ES in place.  

			jmp	@f
@@:
			ret

SetESLimit		ENDP

;   ************************************************************************

END

