
/*  ************************************************************************  *
 *				     main.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <string.h>

#include    <standard.h>

static CHAR description [] = "\n\
Demonstrates behaviour of the descriptor cache for the ES segment\n\
across a transition from protected mode to real mode.\n\
\n\
SEGLIMIT [limit] [/v]\n\
\n\
where the limit is interpreted as hexadecimal and for safety, may not\n\
be < FFFF.\n\
\n\
By default, the program does not execute in a virtual machine, since\n\
even attempting to discover the segment limit may trigger a protection\n\
violation (or worse).  Use the /v switch to override this.\n\
";

extern DWORD GetESLimit (void);
extern void SetESLimit (DWORD);
extern BOOL InProtectedMode (void);
extern BOOL Is80386 (void);

static BOOL ReadLongHex (CHAR *, DWORD *);
static void PutResult (CHAR *, DWORD);
static void PutError (CHAR *, ...);

int main (int argc, char **argv)
{
    register CHAR *ptr;
    DWORD limit;
    BOOL limit_given = FALSE;
    BOOL vm_ok = FALSE;

    /*	Parse the command line, accepting a '/?' switch for a program
	description or a hexadecimal limit for ES.  For the /v switch,
	accept any initial substring of "VirtualMachineAllowed",
	independently of case.	*/

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;
	    if (ptr [0] == '?' AND ptr [1] == '\0') {
		printf (description);
		return (1);
	    }
	    if (strnicmp ("VirtualMachineAllowed", ptr, strlen (ptr)) == 0) {
		vm_ok = TRUE;
		continue;
	    }
	    ptr --;
	    PutError ("invalid switch - %s", ptr);
	    return (-1);
	}
	else {
	    if (limit_given) {
		PutError ("too many parameters");
		return (-1);
	    }
	    if (ReadLongHex (ptr, &limit) AND NOT limit < 0xFFFF) {
		limit_given = TRUE;
		continue;
	    }
	    PutError ("invalid parameter - %s", ptr);
	    return (-1);
	}
    }

    /*	Establish that the 80386-specific functions for manipulating segment
	limits are usable.  */

    if (NOT Is80386 ()) {
	PutError ("requires 80386 CPU or better");
	return (-1);
    }

    if (InProtectedMode ()) {
	if (NOT vm_ok) {
	    PutError ("CPU is presently in protected mode");
	    return (-1);
	}
    }

    /*	Disable interrupts for the test, which may possibly involve setting a
	new limit, then seeing if the operation was successful.  Note that
	segment sizes below 64KB will not be welcomed when interrupts are
	re-enabled.  At present, the program disallows attempts to set
	segment limits < FFFFh, but an alternative would be to test whether
	or not small limits can be set, as long as a reasonable segment
	limit is restored before re-enabling interrupts.  */

    _disable ();
    if (limit_given) SetESLimit (limit);
    limit = GetESLimit ();
    _enable ();

    printf ("\n%s %08lXh\n", 
	limit_given ? "ES limit changed to" : "Current ES limit is", limit);
    return (0);
}

static BOOL ReadLongHex (register CHAR *ptr, DWORD *result)
{
    register WORD ch;
    DWORD temp = 0;

    ch = *ptr;
    if (ch == '0' OR ch == '\\') {
	if ((ptr [1] & ~0x20) == 'X') ptr += 2;
    }
    while (*ptr == '0') ptr ++;

    while (ch = *ptr ++) {
	if (ch >= '0' AND ch <= '9') ch -= '0';
	else {
	    ch &= ~0x20;
	    if (ch >= 'A' AND ch <= 'F') ch -= ('A' - 10);
	    else if (ch == 'H') break;
	    else return (FALSE);
	}
	if (temp & 0xF0000000) return (FALSE);
	temp <<= 4;
	temp += ch;
    }
    *result = temp;
    return (TRUE);
}

static void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nSEGLIMIT fatal error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

