
/*  ************************************************************************  *
 *				   handler.c				      *
 *  ************************************************************************  */

/*  This file contains a sample external A20 handler - close in form to the
    handler HIMEM.SYS provides as AT2.	*/

#include    <conio.h>
#include    <dos.h>

#include    <keepc.h>
#include    <standard.h>

/*  An A20 handler may support up to three functions.  */

#define     DISABLE	0
#define     ENABLE	1
#define     QUERY	2

/*  HIMEM calls the last function (QUERY) only if the ability to service it
    was indicated when the A20 handler's existence was reported via int 2Fh,
    i.e., if the A20DetectionFlag was set to TRUE by InitA20Handler ().

    Although HIMEM's A20 handler for AT machines does not support a query
    facility, keyboard controllers may be able to report the A20 status as
    well as change it.	Compile with the symbol _QUERY_OK_ to include code
    for this purpose.  */

/*  The following pragma makes sure the compiler will generate INP and OUTP
    instructions in-line.  */

#pragma     intrinsic (inp, outp, _disable)

/*  The keyboard controller sets bit 1 at port 64h while it is busy
    processing data written to it.  The following macro generates in-line
    code to wait for the keyboard controller to clear that bit.  */

#define wait_while_busy \
    for (count = 0; inp (0x64) & 0x02 AND -- count; ) {}

/*  Likewise, when data is available from the keyboard controller, it sets
    bit 0.  */

#define wait_for_data	\
    for (count = 0; NOT (inp (0x64) & 0x01) AND -- count; ) {}

BOOL _fastcall _far ExternalA20Handler (int code)
{
    register WORD count;
    int control;

    #ifdef _QUERY_OK_

    if (code == QUERY) {
	_disable ();
	wait_while_busy;
	outp (0x64, 0xD0);		    // Prepare to read output port
	wait_while_busy;
	wait_for_data;
	control = inp (0x60);
	return (control & 0x02 ? TRUE : FALSE);     // Bit 1 is the A20 mask
    }

    #endif

    control = code == DISABLE ? 0xDD : 0xDF;
    _disable ();
    wait_while_busy;
    outp (0x64, 0xD1);			    // Prepare to write output port
    wait_while_busy;
    outp (0x60, control);
    wait_while_busy;
    return (count ? TRUE : FALSE);
}

/*  ************************************************************************  */

