
/*  ************************************************************************  *
 *				   machine.c				      *
 *  ************************************************************************  */

/*  This file contains a (minimal) sample InitA20Handler function, called
    during the XA20 device driver's initialisation.  */

#include    <keepc.h>
#include    <standard.h>

/*  The external A20 handler and an associated flag are defined in the 
    HANDLER.C source file.  */

extern BOOL _fastcall _far ExternalA20Handler (int);
extern BYTE _RESIDENT A20DetectionFlag;

#define     DISABLE	0
#define     ENABLE	1
#define     QUERY	2

/*  The machine for which the external A20 handler is prepared may be
    identified to the user in one of two circumstances:  if the XA20 device
    driver file is executed as a normal DOS program (resulting in a program
    description) or on successful initialisation when actually loaded as a
    device driver.  Default messages will be used in both cases if the
    following variable is a NULL pointer, but in the interests of clarity,
    it is recommended that a description be provided instead.  */

CHAR *pMachineName = "Geoff's cheap 386 clone";

/*  The function called to initialise the external A20 handler must

    1.  set the A20DetectionFlag to TRUE if the external A20 handler
	supports the Query A20 Status function

    2.  physically disable the A20 line if it is already enabled

    in addition to possibly establishing the suitability of the machine on
    which the program is being executed (and/or testing that the proposed
    A20 handler does in fact work).

    Command-line arguments are passed to InitA20Handler for general
    configurability (not taken up in this example).  Standardised error
    messages may be displayed using the PutError or PutWarning functions,
    which follow the style of printf ().  */

extern void PutError (CHAR *, ...);
extern void PutWarning (CHAR *, ...);

BOOL InitA20Handler (int argc, char **argv)
{
    #ifdef _QUERY_OK_
    A20DetectionFlag = TRUE;
    #endif

    ExternalA20Handler (DISABLE);
    return (TRUE);
}

/*  ************************************************************************  */

