
/*  ************************************************************************  *
 *		      main.c - initialisation for XA20.EXE		      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdarg.h>
#include    <stdio.h>

#include    <keepc.h>
#include    <standard.h>

/*  Message strings  */

static CHAR description [] = "\n\
Loads an external A20 handler for subsequent installation by HIMEM.SYS\n\
";

/*  A non-resident routine to test the hardware for suitability and the
    address of a string to describe the configuration - these must be
    provided specially for the machine concerned, but may be NULL (as will
    be the case if they are not linked) if default behaviour is acceptable  */

extern BOOL InitA20Handler (int, char **);
extern CHAR *pMachineName;

/*  The A20 handler in a resident segment - this and an associated flag must
    be provided specially for the machine concerned  */

extern BOOL _fastcall _far ExternalA20Handler (int);

/*  External functions and data concerned with communicating the existence
    of the external A20 handler  */

extern void (_interrupt _far * _RESIDENT lpOrgInt2Fh) (void);
extern void _interrupt _far int2Fh (void);

/*  A function to determine that an XMS server is not yet present  */

extern void (_far *GetXMSAddress (void)) (void);

/*  A function to check that the computer supports an A20 line  */

static BOOL IsAtLeast80286 (void);

/*  A Boolean flag to distinguish normal program execution from device
    driver initialisation  */

extern BOOL _RESIDENT bLoadedAsDriver;

/*  Message output  */

void PutError (CHAR *, ...);
void PutWarning (CHAR *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    /*	Check that the program has been linked with an A20 handler.  */

    if ((void _far *) ExternalA20Handler == NULL) {
	printf (description);
	printf ("\nThis program must be linked with machine-specific code\n");
	return (-1);
    }

    /*	Set a default machine description.  */

    if (pMachineName == NULL) pMachineName = "some un-named machine";

    /*	If an XMS server is already active, then it is too late to load an
	A20 handler for HIMEM to install.  Regard this condition as a fatal
	error if the program is being loaded as a device driver, but
	otherwise, describe the program, taking care to indicate its
	machine-dependent nature.  */

    if (GetXMSAddress ()) {
	if (bLoadedAsDriver) {
	    PutError ("XMS handler already installed");
	    return (-1);
	}
	else {
	    printf (description);
	    printf ("\nThis program has been prepared specifically for %s\n",
		pMachineName);
	    return (0);
	}
    }

    /*	Do not proceed if installation would be clearly pointless.  */

    if (NOT IsAtLeast80286 ()) {
	PutError ("requires 80286 processor or better");
	return (-1);
    }

    /*	If a function for identifying the hardware or checking suitability
	has been linked, call it.  Displaying an error message is the
	responsibility of that function.  */

    if ((void *) InitA20Handler != NULL AND NOT InitA20Handler (argc, argv))
	return (-1);

    /*	Install the int 2Fh service through which HIMEM will discover the
	external A20 handler, then declare success and tidy up.  */

    lpOrgInt2Fh = _dos_getvect (0x2F);
    _dos_setvect (0x2F, int2Fh);

    printf ("\nExternal A20 handler configured for %s\n", pMachineName);

    CloseAllFiles ();
    FreeEnvironment ();
    ExitKeepC (0, 0);
}

static BOOL IsAtLeast80286 (void)
{
    _asm {
	    pushf		    ;  preserve the FLAGS register so
				    ;  that IF can be restored at the end

	    xor     ax,ax	    ;  try clearing all FLAGS bits
	    push    ax
	    popf

	    pushf		    ;  on the 8086 and 80186, bits C to F
	    pop     ax		    ;  are permanently set
	    and     ax,0F000h
	    cmp     ax,0F000h
	    jz	    old
	    mov     ax,0001h
	    jmp     done
	old:
	    xor     ax,ax
	done:
	    popf
    }
}

void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nExternal A20 fatal error:");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

void PutWarning (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nExternal A20 warning:");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

