
/*  ************************************************************************  *
 *				   consult.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <fcntl.h>

#include    <standard.h>

#include    "xmswatch.h"

/*  Functions and data used to detect the presence of a resident copy of
    XMSWATCH and to retrieve the information it may have saved	*/

extern CHAR DeviceName [8];

WORD _dos_getdevicestatus (int, WORD *);
WORD _dos_ioctlread (int, void _far *, WORD, WORD *);

/*  The following function obtains two data structures from the resident
    XMSWATCH copy.  The function's caller need not know that the data is
    obtained via a device driver IOCTL interface, but must understand the
    form in which the data is presented.  The XMSWATCH structure has fixed
    size and the caller must supply the address of some suitable memory.
    Since the LIST structure contains a field of variable length, the caller
    cannot know in advance how much memory to set aside.  Instead, the
    function retrieves the list into memory allocated from the heap and
    returns the address.  */

struct LIST *ConsultWatcher (register struct XMSWATCH *watchptr)
{
    int devicehandle;
    register WORD handle;
    WORD count, status;
    struct LIST *listptr;

    /*	Attempt to open the XMSWWWW0 device.  */

    if (_dos_open (DeviceName, O_RDWR, &devicehandle)) return (FALSE);

    /*	Use DOS function 4400h to verify that _dos_open has indeed found a
	character device (status flags masked by 0x0080) that supports the
	IOCTL interface (mask 0x4000).	*/

    if (_dos_getdevicestatus (handle = devicehandle, &status)
	    OR (status & 0x4080) != 0x4080) goto fail;

    /*	Read the XMSWATCH structure and check that the resident XMSWATCH
	copy supports the expected version of the interface.  */

    if (_dos_ioctlread (handle, watchptr, sizeof (*watchptr), &count)
	    OR count != sizeof (*watchptr)) goto fail;

    if (watchptr -> wVersion != XMSWATCH_VERSION) goto fail;

    /*	Information in the MONITOR structure yields the size of the resident
	handle table.  Ask the device to copy the table into space allocated
	from the near heap.  */

    listptr = CreateList (watchptr -> wHandles);
    if (listptr == NULL) goto fail;

    if (_dos_ioctlread (handle, listptr -> array,
	    watchptr -> wHandleTableSize, &count)
	    OR count != watchptr -> wHandleTableSize) goto fail;

    /*	Close the handle and return a pointer to the dynamically-allocated
	list.  */

    _dos_close (handle);
    return (listptr);

    fail:

    /*	If for any reason, the device is unable to return the expected data,
	close the file handle and return NULL.	The goto statement has been
	used to deal with this cleanup because the alternative involves
	either too many braces and indents or too much repetition.  */

    _dos_close (handle);
    return (NULL);
}

WORD _dos_getdevicestatus (int handle, WORD *statusptr)
{
    WORD status;

    _asm {
	    mov     bx,handle
	    mov     ax,4400h
	    int     21h
	    jc	    fail
	    mov     [status],dx
    }
    *statusptr = status;
    return (0);

    _asm fail:
}

WORD _dos_ioctlread (int handle, void _far *buffer, WORD count, WORD *numread)
{
    _asm {
	    push    ds
	    lds     dx,buffer
	    mov     cx,count
	    mov     bx,handle
	    mov     ax,4402h
	    int     21h
	    pop     ds
	    jc	    fail
	    mov     [count],ax
    }
    *numread = count;
    return (0);

    _asm fail:
}

/*  ************************************************************************  */

