
/*  ************************************************************************  *
 *				   install.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    <driver.h>
#include    <keepc.h>
#include    <psp.h>
#include    <standard.h>
#include    <sysvars.h>

#include    "xmswatch.h"

#pragma intrinsic (_disable, _enable)

/*  A flag provided by the CRTDRVR library to distinguish normal program
    execution from device driver initialisation  */

extern BOOL _RESIDENT bLoadedAsDriver;

/*  The logical name for access to the resident component's device driver
    interface  */

CHAR DeviceName [] = "XMSWWWW0";

/*  The calling point for XMS services - as determined by MAIN.C  */

extern void (_far *lpXMS) (void);

/*  Calling address needed by the resident component  */

extern void (_far * _RESIDENT lpNextXMS) (void);

/*  The address of the handler (in MONITOR.C) that this program adds to
    the XMS chain, a function to perform the addition and a resident
    pre-terminator to activate the handler.  */

extern void _far xms (void);
extern void (_far *ThreadXMSChain (void (_far *) (void))) (void);
extern void _far Activate (void);

/*  The address of DOS's internal variable table (recorded in the resident
    data group by MAIN.C) and of this program's own device driver header, so
    that device driver installation can be mimicked even if the program is
    run as a normal executable.  */

extern struct SYSVARS _far * _RESIDENT lpSysvars;
extern struct DRIVER_HEADER _far DriverHeader;

/*  A function to obtain the number of XMS handles to watch, plus variables
    to describe the tables in which information about these handles will be
    kept.  */

extern WORD QueryFreeHandles (void);

extern WORD _RESIDENT wHandles;
extern struct HANDLE * _RESIDENT pHandleTable;
extern WORD _RESIDENT wHandleTableSize;

/*  A self-explanatory function for fatal message output  */

extern PutError (CHAR *, ...);

/*  ========================================================================  */

int Install (void)
{
    struct find_t fileinfo;
    WORD paras_avail, paras_wanted;

    /*	Disallow an attempt at multiple installation - indicated by the
	prior existence of the XMSWWWW0 device.  Be cautious and do not
	proceed if a file of the same name exists, for installation would
	render that file inaccessible for the remainder of the session.  */

    if (_dos_findfirst (DeviceName,
	    _A_HIDDEN | _A_SYSTEM | _A_SUBDIR, &fileinfo) == 0) {
	PutError ("XMSWWWW0 file or device present already");
	return (-1);
    }

    /*	Determine (and save for the resident component) the number of free
	XMS handles remaining.	It is not possible to obtain via the XMS the
	number of handles that can be provided by the server, so if XMSWATCH
	is not loaded immediately after the primary XMS server, some handles
	may be missed.	*/

    wHandles = QueryFreeHandles ();
    if (wHandles == 0) {
	PutError ("no XMS handles available");
	return (-1);
    }

    /*	Check the availability of space for the table in which the resident
	component will maintain information about XMS handles.	The handle
	table begins at the break address (NOKEEP_BTEXT:0000h) specified by
	the Keep C library.  */

    wHandleTableSize = wHandles * sizeof (struct HANDLE);
    paras_wanted = (wHandleTableSize + 0x0F) >> 4;
    paras_avail = ((struct PSP _far *) MK_FP (_psp, NULL)) -> spCeiling
			- (_segment) lpBreakAddress;

    if (paras_wanted > paras_avail) {
	PutError ("insufficient memory for handle table");
	return (-1);
    }

    /*	Calculate and record the offset address of the handle table with
	respect to the resident data group (the segment address of which is
	held in spKEEP).  */

    pHandleTable = (struct HANDLE *)
		    ((WORD) ((_segment) lpBreakAddress - spKEEP) << 4);

    /*	Thread this program's own XMS handler onto the end of the chain.  */

    lpNextXMS = ThreadXMSChain (xms);
    if (lpNextXMS == NULL) {
	PutError ("bad XMS handler chain");
	return (-1);
    }

    /*	Note that the interception of XMS calls cannot be activated until
	the handle table has been initialised (cleared to zero), but the
	handle table lies in space presently occupied by non-resident code.
	Register the activation routine as a resident pre-terminator to be
	sure it will be called from a safe location.  */

    AtExitKeepC (Activate);

    /*	If the program has not been loaded as a device driver, thread it
	into the device driver chain by hand.  Do not take the chance
	(however small) that an interrupting process will try to follow the
	device driver chain (by hand or through a DOS function) while the
	chain is being adjusted.  */

    if (NOT bLoadedAsDriver) {
	DriverHeader.lpNext = lpSysvars -> NULHeader.lpNext;
	_disable ();
	lpSysvars -> NULHeader.lpNext = &DriverHeader;
	_enable ();
    }

    /*	Display an installation message.  */

    printf ("\nXMSWATCH driver installed - monitoring %d handles\n", wHandles);

    /*	Tidy up, as usual.  Retain enough memory to cover the handle table.  */

    FreeEnvironment ();
    CloseAllFiles ();
    ExitKeepC (0, (BYTE _RESIDENT *) pHandleTable + wHandleTableSize);
}

/*  ************************************************************************  */

