
/*  ************************************************************************  *
 *				   monitor.c				      *
 *  ************************************************************************  */

/*  This is a resident module.	*/

#include    <dos.h>
#include    <memory.h>

#include    <intpack.h>
#include    <standard.h>
#include    <sysvars.h>

#include    "xmswatch.h"

#pragma intrinsic (_disable, _enable, memset)

/*  The XMS entry point at the time of XMSWATCH's installation  */

void (_far *lpOrgXMS) (void) = NULL;

/*  The address at which XMS requests would have been processed (after the
    chain of far jumps) were this module not intercepting them	*/

void (_far *lpNextXMS) (void) = NULL;

/*  Functions that either jump to or call lpNext, but first load register
    values from the structure passed on the stack  */

extern void XMSFilterPass (void);
extern int CallNextXMS (struct REGPACK STACKSEG *);

/*  Items to help identify the DOS program to whom a given address belongs  */

extern CHAR _far * _far GetMemoryOwner (void _far *);
struct SYSVARS _far *lpSysvars = NULL;
WORD _far *lpUMBLinkSeg = NULL;
WORD wDOSVersion = 0;

typedef struct {		    // a ruse to treat an eight character
    CHAR progname [8];		    // program name as a single object
} PROGNAME;			    // for more efficient manipulation

/*  Variables that describe the table in which information about XMS
    allocations is kept  */

WORD wHandles = 0;
struct HANDLE *pHandleTable = 0;
WORD wHandleTableSize = 0;

/*  ========================================================================  */

/*  The handle table is created in memory that was originally used by the
    program's initialisation code.  The Activate function is called from a
    safe location just before the program terminates-and-stays-resident.
    Until the table has been cleared and the bActive flag is set to TRUE,
    all XMS functions are passed transparently to the next handler.  */

static BOOL bActive = FALSE;

void _far Activate (void)
{
    memset (pHandleTable, 0, wHandleTableSize);
    bActive = TRUE;
}

/*  The following function searches the handle table for the entry that
    matches the handle supplied as the function's argument.  It returns the
    address of the corresponding HANDLE structure or NULL if the handle is
    not yet known to the table.  */

static struct HANDLE *GetHandleDescriptor (WORD handle)
{
    register struct HANDLE *hptr;
    register WORD n;

    n = wHandles, hptr = pHandleTable;
    if (n) {
	do {
	    if (hptr -> wHandle == handle) return (hptr);
	} while (++ hptr, -- n);
    }
    return (NULL);
}

/*  ------------------------------------------------------------------------  */

/*  Only two XMS functions are intercepted non-trivially - the ones for
    allocating and releasing an extended memory block.	The objective is to
    record the numerical value of the handle returned by the XMS server so
    that the non-resident component of this program may obtain information
    about the allocation later.  In addition, some attempt is made to
    identify the DOS program requesting the allocation.  */
            
void XMSFilterAllocateEMB (register struct FARPACK STACKSEG *regptr)
{
    register struct HANDLE *hptr;
    CHAR _far *nameptr;

    /*	Do not intercept the function if the handle table has not yet been
	initialised.  */

    if (NOT bActive) XMSFilterPass ();

    /*	XMS functions must be re-entrant.  For present purposes, dodge the
	problem by preventing interrupts:  since allocation requests are
	infrequent, this is not intolerable.  */

    _disable ();

    /*	Call the next handler to deal with the request as the caller
	intends, but examine the result before allowing control to return.  */

    if (CallNextXMS ((struct REGPACK STACKSEG *) regptr)) {

	/*  If the call was successful, find a spare entry in the handle
	    table and record the handle.  */

	hptr = GetHandleDescriptor (0);
	if (hptr != NULL) {
	    hptr -> wHandle = regptr -> dx;

	    /*	Try also to identify the DOS program that sought the XMS
		allocation, copying the program name (eight characters) from
		the DOS memory arena header if this can be located.  */

	    if (nameptr = GetMemoryOwner (regptr -> retaddr))
		*((PROGNAME *) hptr -> cOwner) = *((PROGNAME _far *) nameptr);
	}
    }
}

void XMSFilterFreeEMB (struct FARPACK STACKSEG *regptr)
{
    register WORD handle;
    register struct HANDLE *hptr;

    if (NOT bActive) XMSFilterPass ();

    handle = regptr -> dx;

    _disable ();
    if (CallNextXMS ((struct REGPACK STACKSEG *) regptr)) {

	/*  If a handle in the table has been released, clear the
	    corresponding entry before returning to the XMS caller.  */

	hptr = GetHandleDescriptor (handle);
	if (hptr != NULL) memset (hptr, 0, sizeof (*hptr));
    }
}

/*  ************************************************************************  */

